@@ -4,19 +4,19 @@
  * This program is distributed under the GNU General Public License, version 2.
  * A copy of this license is included with this source.
  *
- * Copyright 2000-2004, Jack Moffitt <jack@xiph.org, 
+ * Copyright 2000-2004, Jack Moffitt <jack@xiph.org,
  *                      Michael Smith <msmith@xiph.org>,
  *                      oddsock <oddsock@xiph.org>,
  *                      Karl Heyes <karl@xiph.org>
  *                      and others (see AUTHORS for details).
- * Copyright 2011-2012, Philipp "ph3-der-loewe" Schafft <lion@lion.leolix.org>,
+ * Copyright 2011-2022, Philipp "ph3-der-loewe" Schafft <lion@lion.leolix.org>,
  */
 
 /* format_mp3.c
-**
-** format plugin for mp3
-**
-*/
+ **
+ ** format plugin for mp3
+ **
+ */
 
 #ifdef HAVE_CONFIG_H
 #include <config.h>
@@ -24,29 +24,26 @@
 
 #include <stdio.h>
 #include <stdlib.h>
+#include <stdbool.h>
 #include <string.h>
 #ifdef HAVE_STRINGS_H
 # include <strings.h>
 #endif
 
+#include "global.h"
 #include "refbuf.h"
 #include "source.h"
 #include "client.h"
+#include "connection.h"
 
 #include "stats.h"
 #include "format.h"
-#include "httpp/httpp.h"
+#include "common/httpp/httpp.h"
 
 #include "logging.h"
 
 #include "format_mp3.h"
 
-#ifdef WIN32
-#define strcasecmp stricmp
-#define strncasecmp strnicmp
-#define snprintf _snprintf
-#endif
-
 #define CATMODULE "format-mp3"
 
 /* Note that this seems to be 8192 in shoutcast - perhaps we want to be the
@@ -61,27 +58,27 @@ static refbuf_t *mp3_get_no_meta (source_t *source);
 static int  format_mp3_create_client_data (source_t *source, client_t *client);
 static void free_mp3_client_data (client_t *client);
 static int format_mp3_write_buf_to_client(client_t *client);
-static void write_mp3_to_file (struct source_tag *source, refbuf_t *refbuf);
+static void write_mp3_to_file (source_t *source, refbuf_t *refbuf);
 static void mp3_set_tag (format_plugin_t *plugin, const char *tag, const char *in_value, const char *charset);
 static void format_mp3_apply_settings(client_t *client, format_plugin_t *format, mount_proxy *mount);
 
 
 typedef struct {
-   unsigned int interval;
-   int metadata_offset;
-   unsigned int since_meta_block;
-   int in_metadata;
-   refbuf_t *associated;
+    unsigned int interval;
+    int metadata_offset;
+    unsigned int since_meta_block;
+    int in_metadata;
+    refbuf_t *associated;
 } mp3_client_data;
 
-int format_mp3_get_plugin (source_t *source)
+int format_mp3_get_plugin(source_t *source)
 {
     const char *metadata;
     format_plugin_t *plugin;
     mp3_state *state = calloc(1, sizeof(mp3_state));
     refbuf_t *meta;
 
-    plugin = (format_plugin_t *)calloc(1, sizeof(format_plugin_t));
+    plugin = (format_plugin_t *) calloc(1, sizeof(format_plugin_t));
 
     plugin->type = FORMAT_TYPE_GENERIC;
     plugin->get_buffer = mp3_get_no_meta;
@@ -92,7 +89,7 @@ int format_mp3_get_plugin (source_t *source)
     plugin->set_tag = mp3_set_tag;
     plugin->apply_settings = format_mp3_apply_settings;
 
-    plugin->contenttype = httpp_getvar (source->parser, "content-type");
+    plugin->contenttype = httpp_getvar(source->parser, "content-type");
     if (plugin->contenttype == NULL) {
         /* We default to MP3 audio for old clients without content types */
         plugin->contenttype = "audio/mpeg";
@@ -118,8 +115,13 @@ int format_mp3_get_plugin (source_t *source)
             state->interval = state->inline_metadata_interval;
         }
     }
+
+    vorbis_comment_init(&plugin->vc);
     source->format = plugin;
-    thread_mutex_create (&state->url_lock);
+    thread_mutex_create(&state->url_lock);
+
+    /* we are in global locked state here, locked by connection_complete_source(). */
+    global.sources_legacy++;
 
     return 0;
 }
@@ -140,35 +142,34 @@ static void mp3_set_tag (format_plugin_t *plugin, const char *tag, const char *i
         return;
     }
 
-    if (in_value)
-    {
-        value = util_conv_string (in_value, charset, plugin->charset);
+    if (in_value) {
+        if (!charset)
+            charset = plugin->charset;
+
+        value = util_conv_string (in_value, charset, "UTF-8");
         if (value == NULL)
-            value = strdup (in_value);
+            value = strdup(in_value);
+    } else {
+        thread_mutex_unlock(&source_mp3->url_lock);
+        return;
     }
 
-    if (strcmp (tag, "title") == 0 || strcmp (tag, "song") == 0)
-    {
-        free (source_mp3->url_title);
-        source_mp3->url_title = value;
+    if (strcmp(tag, "title") == 0 || strcmp(tag, "song") == 0) {
+        tag = MP3_METADATA_TITLE;
+    } else if (strcmp(tag, "artist") == 0) {
+        tag = MP3_METADATA_ARTIST;
+    } else if (strcmp(tag, "url") == 0) {
+        tag = MP3_METADATA_URL;
     }
-    else if (strcmp (tag, "artist") == 0)
-    {
-        free (source_mp3->url_artist);
-        source_mp3->url_artist = value;
-    }
-    else if (strcmp (tag, "url") == 0)
-    {
-        free (source_mp3->url);
-        source_mp3->url = value;
-    }
-    else
-        free (value);
+
+    format_set_vorbiscomment(plugin, tag, value);
+    free (value);
+
     thread_mutex_unlock (&source_mp3->url_lock);
 }
 
 
-static void filter_shoutcast_metadata (source_t *source, char *metadata, unsigned int meta_len)
+static void filter_shoutcast_metadata (source_t *source, char *metadata, unsigned int meta_len, bool update_vc)
 {
     if (metadata)
     {
@@ -189,8 +190,12 @@ static void filter_shoutcast_metadata (source_t *source, char *metadata, unsigne
                 memcpy (p, metadata+13, len);
                 logging_playlist (source->mount, p, source->listeners);
                 stats_event_conv (source->mount, "title", p, source->format->charset);
+                stats_event_conv (source->mount, "display-title", p, source->format->charset);
+                if (update_vc)
+                    format_set_vorbiscomment(source->format, MP3_METADATA_TITLE, p);
                 yp_touch (source->mount);
                 free (p);
+                playlist_push_track(source->history, &source->format->vc);
             }
         } while (0);
     }
@@ -224,8 +229,10 @@ static void format_mp3_apply_settings (client_t *client, format_plugin_t *format
         }
     }
 
-    if (format->charset == NULL)
-        format->charset = strdup ("ISO8859-1");
+    if (format->charset == NULL) {
+        ICECAST_LOG_INFO("No charset given for mount %#H with source client %zu, assuming ISO8859-1", mount ? mount->mountname : NULL, client->con->id);
+        format->charset = strdup("ISO8859-1");
+    }
 
     ICECAST_LOG_DEBUG("sending metadata interval %d", source_mp3->interval);
     ICECAST_LOG_DEBUG("charset %s", format->charset);
@@ -235,39 +242,60 @@ static void format_mp3_apply_settings (client_t *client, format_plugin_t *format
 /* called from the source thread when the metadata has been updated.
  * The artist title are checked and made ready for clients to send
  */
-static void mp3_set_title (source_t *source)
+static void mp3_set_title(source_t *source)
 {
     const char streamtitle[] = "StreamTitle='";
     const char streamurl[] = "StreamUrl='";
+    const char *in_url_artist = vorbis_comment_query(&source->format->vc, MP3_METADATA_ARTIST, 0);
+    const char *in_url_title = vorbis_comment_query(&source->format->vc, MP3_METADATA_TITLE, 0);
+    const char *in_url = vorbis_comment_query(&source->format->vc, MP3_METADATA_URL, 0);
+    char *url_artist = NULL;
+    char *url_title = NULL;
+    char *url = NULL;
     size_t size;
     unsigned char len_byte;
     refbuf_t *p;
     unsigned int len = sizeof(streamtitle) + 2; /* the StreamTitle, quotes, ; and null */
     mp3_state *source_mp3 = source->format->_state;
+    const char *charset = source->format->charset;
 
     /* make sure the url data does not disappear from under us */
     thread_mutex_lock (&source_mp3->url_lock);
 
     /* work out message length */
-    if (source_mp3->url_artist)
-        len += strlen (source_mp3->url_artist);
-    if (source_mp3->url_title)
-        len += strlen (source_mp3->url_title);
-    if (source_mp3->url_artist && source_mp3->url_title)
+    if (in_url_artist) {
+        url_artist = util_conv_string(in_url_artist, "UTF-8", charset);
+        if (url_artist)
+            len += strlen(url_artist);
+    }
+
+    if (in_url_title) {
+        url_title = util_conv_string(in_url_title, "UTF-8", charset);
+        if (url_title)
+            len += strlen(url_title);
+    }
+
+    if (url_artist && url_title)
         len += 3;
-    if (source_mp3->inline_url)
-    {
+
+    if (source_mp3->inline_url) {
         char *end = strstr (source_mp3->inline_url, "';");
         if (end)
             len += end - source_mp3->inline_url+2;
+    } else if (in_url) {
+        url = util_conv_string(in_url, "UTF-8", charset);
+        if (url)
+            len += strlen(url) + strlen(streamurl) + 2;
     }
-    else if (source_mp3->url)
-        len += strlen (source_mp3->url) + strlen (streamurl) + 2;
+
 #define MAX_META_LEN 255*16
     if (len > MAX_META_LEN)
     {
         thread_mutex_unlock (&source_mp3->url_lock);
         ICECAST_LOG_WARN("Metadata too long at %d chars", len);
+        free(url_artist);
+        free(url_title);
+        free(url);
         return;
     }
     /* work out the metadata len byte */
@@ -283,12 +311,16 @@ static void mp3_set_title (source_t *source)
         int r;
 
         memset (p->data, '\0', size);
-        if (source_mp3->url_artist && source_mp3->url_title)
+        if (url_artist && url_title) {
             r = snprintf (p->data, size, "%c%s%s - %s';", len_byte, streamtitle,
-                    source_mp3->url_artist, source_mp3->url_title);
-        else
+                    url_artist, url_title);
+        } else if (url_title) {
             r = snprintf (p->data, size, "%c%s%s';", len_byte, streamtitle,
-                    source_mp3->url_title);
+                    url_title);
+        } else {
+            r = snprintf (p->data, size, "%c%s';", len_byte, streamtitle);
+        }
+
         if (r > 0)
         {
             if (source_mp3->inline_url)
@@ -299,16 +331,20 @@ static void mp3_set_title (source_t *source)
                 if ((ssize_t)(size-r) > urllen)
                     snprintf (p->data+r, size-r, "StreamUrl='%s';", source_mp3->inline_url+11);
             }
-            else if (source_mp3->url)
-                snprintf (p->data+r, size-r, "StreamUrl='%s';", source_mp3->url);
+            else if (url)
+                snprintf (p->data+r, size-r, "StreamUrl='%s';", url);
         }
         ICECAST_LOG_DEBUG("shoutcast metadata block setup with %s", p->data+1);
-        filter_shoutcast_metadata (source, p->data, size);
+        filter_shoutcast_metadata (source, p->data, size, false);
 
         refbuf_release (source_mp3->metadata);
         source_mp3->metadata = p;
     }
     thread_mutex_unlock (&source_mp3->url_lock);
+
+    free(url_artist);
+    free(url_title);
+    free(url);
 }
 
 
@@ -345,7 +381,7 @@ static int send_stream_metadata (client_t *client, refbuf_t *associated)
             meta_len = 17 - client_mp3->metadata_offset;
         }
     }
-    ret = client_send_bytes (client, metadata, meta_len);
+    ret = client_send_bytes(client, metadata, meta_len);
 
     if (ret == meta_len)
     {
@@ -382,7 +418,7 @@ static int format_mp3_write_buf_to_client(client_t *client)
         if (client_mp3->in_metadata)
         {
             refbuf_t *associated = refbuf->associated;
-            ret = send_stream_metadata (client, associated);
+            ret = send_stream_metadata(client, associated);
 
             if (client_mp3->in_metadata)
                 break;
@@ -449,25 +485,28 @@ static void format_mp3_free_plugin(format_plugin_t *self)
     /* free the plugin instance */
     mp3_state *state = self->_state;
 
-    thread_mutex_destroy (&state->url_lock);
-    free (state->url_artist);
-    free (state->url_title);
-    free (self->charset);
-    refbuf_release (state->metadata);
-    refbuf_release (state->read_data);
+    thread_mutex_destroy(&state->url_lock);
+    free(self->charset);
+    refbuf_release(state->metadata);
+    refbuf_release(state->read_data);
     free(state);
+    vorbis_comment_clear(&self->vc);
     free(self);
+
+    global_lock();
+    global.sources_legacy--;
+    global_unlock();
 }
 
 
 /* This does the actual reading, making sure the read data is packaged in
  * blocks of 1400 bytes (near the common MTU size). This is because many
- * incoming streams come in small packets which could waste a lot of 
+ * incoming streams come in small packets which could waste a lot of
  * bandwidth with many listeners due to headers and such like.
  */
-static int complete_read (source_t *source)
+static int complete_read(source_t *source)
 {
-    int bytes;
+    ssize_t bytes;
     format_plugin_t *format = source->format;
     mp3_state *source_mp3 = format->_state;
     char *buf;
@@ -477,15 +516,16 @@ static int complete_read (source_t *source)
 
     if (source_mp3->read_data == NULL)
     {
-        source_mp3->read_data = refbuf_new (REFBUF_SIZE); 
+        source_mp3->read_data = refbuf_new (REFBUF_SIZE);
         source_mp3->read_count = 0;
     }
     buf = source_mp3->read_data->data + source_mp3->read_count;
 
-    bytes = client_read_bytes (source->client, buf, REFBUF_SIZE-source_mp3->read_count);
+    bytes = client_body_read(source->client, buf, REFBUF_SIZE-source_mp3->read_count);
     if (bytes < 0)
     {
-        if (source->client->con->error)
+        /* Why do we do this here (not source.c)? -- ph3-der-loewe, 2018-04-17 */
+        if (client_body_eof(source->client))
         {
             refbuf_release (source_mp3->read_data);
             source_mp3->read_data = NULL;
@@ -538,7 +578,7 @@ static refbuf_t *mp3_get_no_meta (source_t *source)
  * metadata so that the mp3 data itself is store on the queue and the
  * metadata is is associated with it
  */
-static refbuf_t *mp3_get_filter_meta (source_t *source)
+static refbuf_t *mp3_get_filter_meta(source_t *source)
 {
     refbuf_t *refbuf;
     format_plugin_t *plugin = source->format;
@@ -604,7 +644,7 @@ static refbuf_t *mp3_get_filter_meta (source_t *source)
             source_mp3->build_metadata_offset += bytes;
             break;
         }
-        /* copy all bytes except the last one, that way we 
+        /* copy all bytes except the last one, that way we
          * know a null byte terminates the message */
         memcpy (source_mp3->build_metadata + source_mp3->build_metadata_offset,
                 src, metadata_remaining-1);
@@ -621,11 +661,11 @@ static refbuf_t *mp3_get_filter_meta (source_t *source)
             memcpy (meta->data, source_mp3->build_metadata,
                     source_mp3->build_metadata_len);
 
-	    ICECAST_LOG_DEBUG("shoutcast metadata %.*s", 4080, meta->data+1);
+            ICECAST_LOG_DEBUG("shoutcast metadata %.*s", 4080, meta->data+1);
             if (strncmp (meta->data+1, "StreamTitle=", 12) == 0)
             {
                 filter_shoutcast_metadata (source, source_mp3->build_metadata,
-                        source_mp3->build_metadata_len);
+                        source_mp3->build_metadata_len, true);
                 refbuf_release (source_mp3->metadata);
                 source_mp3->metadata = meta;
                 source_mp3->inline_url = strstr (meta->data+1, "StreamUrl='");
@@ -658,7 +698,7 @@ static refbuf_t *mp3_get_filter_meta (source_t *source)
 
 static int format_mp3_create_client_data(source_t *source, client_t *client)
 {
-    mp3_client_data *client_mp3 = calloc(1,sizeof(mp3_client_data));
+    mp3_client_data *client_mp3 = calloc(1, sizeof(mp3_client_data));
     mp3_state *source_mp3 = source->format->_state;
     const char *metadata;
     /* the +-2 is for overwriting the last set of \r\n */
@@ -718,7 +758,7 @@ static void free_mp3_client_data (client_t *client)
 }
 
 
-static void write_mp3_to_file (struct source_tag *source, refbuf_t *refbuf)
+static void write_mp3_to_file (source_t *source, refbuf_t *refbuf)
 {
     if (refbuf->len == 0)
         return;
