@@ -3,12 +3,12 @@
  * This program is distributed under the GNU General Public License, version 2.
  * A copy of this license is included with this source.
  *
- * Copyright 2000-2004, Jack Moffitt <jack@xiph.org, 
+ * Copyright 2000-2004, Jack Moffitt <jack@xiph.org,
  *                      Michael Smith <msmith@xiph.org>,
  *                      oddsock <oddsock@xiph.org>,
  *                      Karl Heyes <karl@xiph.org>
  *                      and others (see AUTHORS for details).
- * Copyright 2011,      Philipp "ph3-der-loewe" Schafft <lion@lion.leolix.org>.
+ * Copyright 2011-2018, Philipp "ph3-der-loewe" Schafft <lion@lion.leolix.org>.
  */
 
 #ifdef HAVE_CONFIG_H
@@ -23,7 +23,7 @@
 #include <errno.h>
 
 #ifdef HAVE_POLL
-#include <sys/poll.h>
+#include <poll.h>
 #endif
 
 #ifndef _WIN32
@@ -35,34 +35,31 @@
 #else
 #include <winsock2.h>
 #include <windows.h>
-#define fseeko fseek
 #define SCN_OFF_T "ld"
 #define PRI_OFF_T "ld"
-#define snprintf _snprintf
-#define strncasecmp _strnicmp
 #ifndef S_ISREG
 #define S_ISREG(mode)  ((mode) & _S_IFREG)
 #endif
 #endif
 
-#include "thread/thread.h"
-#include "avl/avl.h"
-#include "httpp/httpp.h"
-#include "net/sock.h"
+#include "common/thread/thread.h"
+#include "common/avl/avl.h"
+#include "common/httpp/httpp.h"
+#include "common/net/sock.h"
 
+#include "fserve.h"
+#include "compat.h"
 #include "connection.h"
 #include "global.h"
 #include "refbuf.h"
 #include "client.h"
+#include "errors.h"
 #include "stats.h"
 #include "format.h"
 #include "logging.h"
 #include "cfgfile.h"
 #include "util.h"
 #include "admin.h"
-#include "compat.h"
-
-#include "fserve.h"
 
 #undef CATMODULE
 #define CATMODULE "fserve"
@@ -79,7 +76,7 @@ static avl_tree *mimetypes = NULL;
 
 static volatile int run_fserv = 0;
 static unsigned int fserve_clients;
-static int client_tree_changed=0;
+static int client_tree_changed = 0;
 
 #ifdef HAVE_POLL
 static struct pollfd *ufds = NULL;
@@ -151,55 +148,57 @@ int fserve_client_waiting (void)
     unsigned int i = 0;
 
     /* only rebuild ufds if there are clients added/removed */
-    if (client_tree_changed)
-    {
-        client_tree_changed = 0;
-        ufds = realloc(ufds, fserve_clients * sizeof(struct pollfd));
-        fclient = active_list;
-        while (fclient)
-        {
-            ufds[i].fd = fclient->client->con->sock;
-            ufds[i].events = POLLOUT;
-            ufds[i].revents = 0;
-            fclient = fclient->next;
-            i++;
+    if (client_tree_changed) {
+        struct pollfd *ufds_new = realloc(ufds, fserve_clients * sizeof(struct pollfd));
+        /* REVIEW: If we can not allocate new ufds, keep old ones for now. */
+        if (ufds_new || fserve_clients == 0) {
+            ufds = ufds_new;
+            client_tree_changed = 0;
+            fclient = active_list;
+            while (fclient)
+            {
+                ufds[i].fd = fclient->client->con->sock;
+                ufds[i].events = POLLOUT;
+                ufds[i].revents = 0;
+                fclient = fclient->next;
+                i++;
+            }
         }
     }
-    if (!ufds)
-    {
+
+    if (!ufds) {
         thread_spin_lock (&pending_lock);
         run_fserv = 0;
         thread_spin_unlock (&pending_lock);
         return -1;
-    }
-    else if (poll(ufds, fserve_clients, 200) > 0)
-    {
+    } else if (poll(ufds, fserve_clients, 200) > 0) {
         /* mark any clients that are ready */
         fclient = active_list;
         for (i=0; i<fserve_clients; i++)
         {
             if (ufds[i].revents & (POLLOUT|POLLHUP|POLLERR))
-                fclient->ready = 1;
+            fclient->ready = 1;
             fclient = fclient->next;
         }
         return 1;
     }
+
     return 0;
 }
 #else
-int fserve_client_waiting (void)
+int fserve_client_waiting(void)
 {
     fserve_t *fclient;
     fd_set realfds;
 
     /* only rebuild fds if there are clients added/removed */
-    if(client_tree_changed) {
+    if (client_tree_changed) {
         client_tree_changed = 0;
         FD_ZERO(&fds);
         fd_max = SOCK_ERROR;
         fclient = active_list;
         while (fclient) {
-            FD_SET (fclient->client->con->sock, &fds);
+            FD_SET(fclient->client->con->sock, &fds);
             if (fclient->client->con->sock > fd_max || fd_max == SOCK_ERROR)
                 fd_max = fclient->client->con->sock;
             fclient = fclient->next;
@@ -261,7 +260,7 @@ static int wait_for_fds(void)
                 fserve_clients++;
             }
             pending_list = NULL;
-            thread_spin_unlock (&pending_lock);
+            thread_spin_unlock(&pending_lock);
         }
         /* drop out of here if someone is ready */
         ret = fserve_client_waiting();
@@ -276,6 +275,8 @@ static void *fserv_thread_function(void *arg)
     fserve_t *fclient, **trail;
     size_t bytes;
 
+    (void)arg;
+
     while (1)
     {
         if (wait_for_fds() < 0)
@@ -344,7 +345,7 @@ static void *fserv_thread_function(void *arg)
 }
 
 /* string returned needs to be free'd */
-char *fserve_content_type (const char *path)
+char *fserve_content_type(const char *path)
 {
     char *ext = util_get_extension(path);
     mime_type exttype = {ext, NULL};
@@ -404,7 +405,7 @@ static void fserve_client_destroy(fserve_t *fclient)
 /* client has requested a file, so check for it and send the file.  Do not
  * refer to the client_t afterwards.  return 0 for success, -1 on error.
  */
-int fserve_client_create (client_t *httpclient, const char *path)
+int fserve_client_create (client_t *httpclient)
 {
     int bytes;
     struct stat file_buf;
@@ -420,8 +421,8 @@ int fserve_client_create (client_t *httpclient, const char *path)
     ice_config_t *config;
     FILE *file;
 
-    fullpath = util_get_path_from_normalised_uri (path);
-    ICECAST_LOG_INFO("checking for file %H (%H)", path, fullpath);
+    fullpath = util_get_path_from_normalised_uri(httpclient->uri);
+    ICECAST_LOG_INFO("checking for file %H (%H)", httpclient->uri, fullpath);
 
     if (strcmp (util_get_extension (fullpath), "m3u") == 0)
         m3u_requested = 1;
@@ -439,7 +440,7 @@ int fserve_client_create (client_t *httpclient, const char *path)
         if (m3u_requested == 0 && xslt_playlist_requested == NULL)
         {
             ICECAST_LOG_WARN("req for file \"%H\" %s", fullpath, strerror (errno));
-            client_send_404 (httpclient, "The file you requested could not be found");
+            client_send_error_by_id(httpclient, ICECAST_ERROR_FSERV_FILE_NOT_FOUND);
             free (fullpath);
             return -1;
         }
@@ -451,44 +452,21 @@ int fserve_client_create (client_t *httpclient, const char *path)
 
     if (m3u_requested && m3u_file_available == 0)
     {
-        const char *host = httpp_getvar (httpclient->parser, "host");
-        char *sourceuri = strdup (path);
+        char *sourceuri = strdup(httpclient->uri);
         char *dot = strrchr(sourceuri, '.');
 
-        /* at least a couple of players (fb2k/winamp) are reported to send a 
-         * host header but without the port number. So if we are missing the
-         * port then lets treat it as if no host line was sent */
-        if (host && strchr (host, ':') == NULL)
-            host = NULL;
-
         *dot = 0;
         httpclient->respcode = 200;
         ret = util_http_build_header (httpclient->refbuf->data, BUFSIZE, 0,
-	                              0, 200, NULL,
-				      "audio/x-mpegurl", NULL, "", NULL);
+                                      0, 200, NULL,
+                                      "audio/x-mpegurl", NULL, "", NULL, httpclient);
         if (ret == -1 || ret >= (BUFSIZE - 512)) { /* we want at least 512 bytes left for the content of the playlist */
             ICECAST_LOG_ERROR("Dropping client as we can not build response headers.");
-            client_send_500(httpclient, "Header generation failed.");
+            client_send_error_by_id(httpclient, ICECAST_ERROR_GEN_HEADER_GEN_FAILED);
+            free(sourceuri);
             return -1;
         }
-        if (host == NULL)
-        {
-	    config = config_get_config();
-            snprintf (httpclient->refbuf->data + ret, BUFSIZE - ret,
-                    "http://%s:%d%s\r\n", 
-                    config->hostname, config->port,
-                    sourceuri
-                    );
-            config_release_config();
-        }
-        else
-        {
-	    snprintf (httpclient->refbuf->data + ret, BUFSIZE - ret,
-                    "http://%s%s\r\n", 
-                    host, 
-                    sourceuri
-                    );
-        }
+        client_get_baseurl(httpclient, NULL, httpclient->refbuf->data + ret, BUFSIZE - ret, NULL, NULL, NULL, sourceuri, "\r\n");
         httpclient->refbuf->len = strlen (httpclient->refbuf->data);
         fserve_add_client (httpclient, NULL);
         free (sourceuri);
@@ -498,14 +476,15 @@ int fserve_client_create (client_t *httpclient, const char *path)
     if (xslt_playlist_requested && xslt_playlist_file_available == 0)
     {
         xmlDocPtr doc;
-        char *reference = strdup (path);
+        char *reference = strdup(httpclient->uri);
         char *eol = strrchr (reference, '.');
         if (eol)
             *eol = '\0';
-        doc = stats_get_xml (0, reference);
+        doc = stats_get_xml(STATS_XML_FLAG_NONE, reference, httpclient);
         free (reference);
-        admin_send_response (doc, httpclient, TRANSFORMED, xslt_playlist_requested);
+        admin_send_response (doc, httpclient, ADMIN_FORMAT_HTML, xslt_playlist_requested);
         xmlFreeDoc(doc);
+        free (fullpath);
         return 0;
     }
 
@@ -514,16 +493,16 @@ int fserve_client_create (client_t *httpclient, const char *path)
     if (config->fileserve == 0)
     {
         ICECAST_LOG_DEBUG("on demand file \"%H\" refused. Serving static files has been disabled in the config", fullpath);
-        client_send_404 (httpclient, "The file you requested could not be found");
+        client_send_error_by_id(httpclient, ICECAST_ERROR_FSERV_FILE_NOT_FOUND);
         config_release_config();
-        free (fullpath);
+        free(fullpath);
         return -1;
     }
     config_release_config();
 
     if (S_ISREG (file_buf.st_mode) == 0)
     {
-        client_send_404 (httpclient, "The file you requested could not be found");
+        client_send_error_by_id(httpclient, ICECAST_ERROR_FSERV_FILE_NOT_FOUND);
         ICECAST_LOG_WARN("found requested file but there is no handler for it: %H", fullpath);
         free (fullpath);
         return -1;
@@ -533,7 +512,7 @@ int fserve_client_create (client_t *httpclient, const char *path)
     if (file == NULL)
     {
         ICECAST_LOG_WARN("Problem accessing file \"%H\"", fullpath);
-        client_send_404 (httpclient, "File not readable");
+        client_send_error_by_id(httpclient, ICECAST_ERROR_FSERV_FILE_NOT_READABLE);
         free (fullpath);
         return -1;
     }
@@ -575,14 +554,14 @@ int fserve_client_create (client_t *httpclient, const char *path)
                     endpos = 0;
                 }
                 httpclient->respcode = 206;
-                type = fserve_content_type (path);
-		bytes = util_http_build_header (httpclient->refbuf->data, BUFSIZE, 0,
-		                                0, 206, NULL,
-						type, NULL,
-						NULL, NULL);
+                type = fserve_content_type(httpclient->uri);
+                bytes = util_http_build_header (httpclient->refbuf->data, BUFSIZE, 0,
+                                                0, 206, NULL,
+                                                type, NULL,
+                                                NULL, NULL, httpclient);
                 if (bytes == -1 || bytes >= (BUFSIZE - 512)) { /* we want at least 512 bytes left */
                     ICECAST_LOG_ERROR("Dropping client as we can not build response headers.");
-                    client_send_500(httpclient, "Header generation failed.");
+                    client_send_error_by_id(httpclient, ICECAST_ERROR_GEN_HEADER_GEN_FAILED);
                     return -1;
                 }
                 bytes += snprintf (httpclient->refbuf->data + bytes, BUFSIZE - bytes,
@@ -605,15 +584,16 @@ int fserve_client_create (client_t *httpclient, const char *path)
         }
     }
     else {
-        char *type = fserve_content_type(path);
+        char *type = fserve_content_type(httpclient->uri);
         httpclient->respcode = 200;
-	bytes = util_http_build_header (httpclient->refbuf->data, BUFSIZE, 0,
-	                                0, 200, NULL,
-					type, NULL,
-					NULL, NULL);
+        bytes = util_http_build_header (httpclient->refbuf->data, BUFSIZE, 0,
+                                        0, 200, NULL,
+                                        type, NULL,
+                                        NULL, NULL, httpclient);
         if (bytes == -1 || bytes >= (BUFSIZE - 512)) { /* we want at least 512 bytes left */
             ICECAST_LOG_ERROR("Dropping client as we can not build response headers.");
-            client_send_500(httpclient, "Header generation failed.");
+            client_send_error_by_id(httpclient, ICECAST_ERROR_GEN_HEADER_GEN_FAILED);
+            fclose(file);
             return -1;
         }
         bytes += snprintf (httpclient->refbuf->data + bytes, BUFSIZE - bytes,
@@ -632,7 +612,7 @@ int fserve_client_create (client_t *httpclient, const char *path)
 
 fail:
     fclose (file);
-    client_send_error(httpclient, 416, 1, "Request Range Not Satisfiable\r\n");
+    client_send_error_by_id(httpclient, ICECAST_ERROR_FSERV_REQUEST_RANGE_NOT_SATISFIABLE);
     return -1;
 }
 
@@ -662,10 +642,10 @@ int fserve_add_client (client_t *client, FILE *file)
 {
     fserve_t *fclient = calloc (1, sizeof(fserve_t));
 
-    ICECAST_LOG_DEBUG("Adding client to file serving engine");
+    ICECAST_LOG_DEBUG("Adding client %p to file serving engine", client);
     if (fclient == NULL)
     {
-        client_send_404 (client, "memory exhausted");
+        client_send_error_by_id(client, ICECAST_ERROR_GEN_MEMORY_EXHAUSTED);
         return -1;
     }
     fclient->file = file;
@@ -687,7 +667,7 @@ void fserve_add_client_callback (client_t *client, fserve_callback_t callback, v
     ICECAST_LOG_DEBUG("Adding client to file serving engine");
     if (fclient == NULL)
     {
-        client_send_404 (client, "memory exhausted");
+        client_send_error_by_id(client, ICECAST_ERROR_GEN_MEMORY_EXHAUSTED);
         return;
     }
     fclient->file = NULL;
@@ -696,7 +676,7 @@ void fserve_add_client_callback (client_t *client, fserve_callback_t callback, v
     fclient->callback = callback;
     fclient->arg = arg;
 
-    fserve_add_pending (fclient);
+    fserve_add_pending(fclient);
 }
 
 
@@ -711,12 +691,13 @@ static int _delete_mapping(void *mapping) {
 
 static int _compare_mappings(void *arg, void *a, void *b)
 {
+    (void)arg;
     return strcmp(
             ((mime_type *)a)->ext,
             ((mime_type *)b)->ext);
 }
 
-void fserve_recheck_mime_types (ice_config_t *config)
+void fserve_recheck_mime_types(ice_config_t *config)
 {
     FILE *mimefile;
     char line[4096];
