@@ -3,11 +3,12 @@
  * This program is distributed under the GNU General Public License, version 2.
  * A copy of this license is included with this source.
  *
- * Copyright 2000-2004, Jack Moffitt <jack@xiph.org, 
+ * Copyright 2000-2004, Jack Moffitt <jack@xiph.org,
  *                      Michael Smith <msmith@xiph.org>,
  *                      oddsock <oddsock@xiph.org>,
  *                      Karl Heyes <karl@xiph.org>
  *                      and others (see AUTHORS for details).
+ * Copyright 2012-2022, Philipp "ph3-der-loewe" Schafft <lion@lion.leolix.org>,
  */
 
 /* -*- c-basic-offset: 4; indent-tabs-mode: nil; -*- */
@@ -22,6 +23,7 @@
 #include <ogg/ogg.h>
 #include <errno.h>
 
+/* REVIEW: Are all those includes needed? */
 #ifndef _WIN32
 #include <unistd.h>
 #include <sys/time.h>
@@ -37,30 +39,28 @@
 #define snprintf _snprintf
 #endif
 
-#ifndef _WIN32
-/* for __setup_empty_script_environment() */
-#include <sys/stat.h>
-#include <fcntl.h>
-#endif
-
-#include "thread/thread.h"
-#include "avl/avl.h"
-#include "httpp/httpp.h"
-#include "net/sock.h"
+#include "common/thread/thread.h"
+#include "common/avl/avl.h"
+#include "common/httpp/httpp.h"
 
+#include "source.h"
+#include "compat.h"
 #include "connection.h"
 #include "global.h"
 #include "refbuf.h"
 #include "client.h"
+#include "errors.h"
 #include "stats.h"
 #include "logging.h"
 #include "cfgfile.h"
 #include "util.h"
-#include "source.h"
 #include "format.h"
 #include "fserve.h"
 #include "auth.h"
-#include "compat.h"
+#include "event.h"
+#include "slave.h"
+#include "acl.h"
+#include "navigation.h"
 
 #undef CATMODULE
 #define CATMODULE "source"
@@ -70,15 +70,9 @@
 mutex_t move_clients_mutex;
 
 /* avl tree helper */
-static int _compare_clients(void *compare_arg, void *a, void *b);
 static int _free_client(void *key);
 static void _parse_audio_info (source_t *source, const char *s);
 static void source_shutdown (source_t *source);
-#ifdef _WIN32
-#define source_run_script(x,y)  ICECAST_LOG_WARN("on [dis]connect scripts disabled");
-#else
-static void source_run_script (char *command, char *mountpoint);
-#endif
 
 /* Allocate a new source with the stated mountpoint, if one already
  * exists with that mountpoint in the global source tree then return
@@ -106,19 +100,22 @@ source_t *source_reserve (const char *mount)
         if (src == NULL)
             break;
 
-        src->client_tree = avl_tree_new(_compare_clients, NULL);
-        src->pending_tree = avl_tree_new(_compare_clients, NULL);
+        src->client_tree = avl_tree_new(client_compare, NULL);
+        src->pending_tree = avl_tree_new(client_compare, NULL);
+        src->history = playlist_new(10 /* DOCUMENT: default is max_tracks=10. */);
 
         /* make duplicates for strings or similar */
-        src->mount = strdup (mount);
+        src->mount = strdup(mount);
+        src->identifier = mount_identifier_new(mount);
         src->max_listeners = -1;
+        src->allow_direct_access = true;
         thread_mutex_create(&src->lock);
 
-        avl_insert (global.source_tree, src);
+        avl_insert(global.source_tree, src);
 
     } while (0);
 
-    avl_tree_unlock (global.source_tree);
+    avl_tree_unlock(global.source_tree);
     return src;
 }
 
@@ -137,18 +134,18 @@ source_t *source_find_mount_raw(const char *mount)
     }
     /* get the root node */
     node = global.source_tree->root->right;
-    
+
     while (node) {
-        source = (source_t *)node->key;
+        source = (source_t *) node->key;
         cmp = strcmp(mount, source->mount);
-        if (cmp < 0) 
+        if (cmp < 0)
             node = node->left;
         else if (cmp > 0)
             node = node->right;
         else
             return source;
     }
-    
+
     /* didn't find it */
     return NULL;
 }
@@ -158,7 +155,7 @@ source_t *source_find_mount_raw(const char *mount)
  * check the fallback, and so on.  Must have a global source lock to call
  * this function.
  */
-source_t *source_find_mount (const char *mount)
+source_t *source_find_mount_with_history(const char *mount, navigation_history_t *history)
 {
     source_t *source = NULL;
     ice_config_t *config;
@@ -170,16 +167,26 @@ source_t *source_find_mount (const char *mount)
     {
         source = source_find_mount_raw(mount);
 
-        if (source)
-        {
+        if (source) {
+            if (history)
+                navigation_history_navigate_to(history, source->identifier, NAVIGATION_DIRECTION_DOWN);
+
             if (source->running || source->on_demand)
                 break;
+        } else {
+            if (history) {
+                mount_identifier_t *identifier = mount_identifier_new(mount);
+                if (identifier) {
+                    navigation_history_navigate_to(history, identifier, NAVIGATION_DIRECTION_DOWN);
+                    refobject_unref(identifier);
+                }
+            }
         }
 
         /* we either have a source which is not active (relay) or no source
          * at all. Check the mounts list for fallback settings
          */
-        mountinfo = config_find_mount (config, mount, MOUNT_TYPE_NORMAL);
+        mountinfo = config_find_mount(config, mount, MOUNT_TYPE_NORMAL);
         source = NULL;
 
         if (mountinfo == NULL)
@@ -192,12 +199,13 @@ source_t *source_find_mount (const char *mount)
     return source;
 }
 
-
 int source_compare_sources(void *arg, void *a, void *b)
 {
     source_t *srca = (source_t *)a;
     source_t *srcb = (source_t *)b;
 
+    (void)arg;
+
     return strcmp(srca->mount, srcb->mount);
 }
 
@@ -282,7 +290,7 @@ void source_clear_source (source_t *source)
     source->hidden = 0;
     source->shoutcast_compat = 0;
     source->client_stats_update = 0;
-    util_dict_free (source->audio_info);
+    util_dict_free(source->audio_info);
     source->audio_info = NULL;
 
     free(source->fallback_mount);
@@ -291,6 +299,9 @@ void source_clear_source (source_t *source)
     free(source->dumpfilename);
     source->dumpfilename = NULL;
 
+    playlist_release(source->history);
+    source->history = NULL;
+
     if (source->intro_file)
     {
         fclose (source->intro_file);
@@ -316,6 +327,7 @@ void source_free_source (source_t *source)
     /* make sure all YP entries have gone */
     yp_remove (source->mount);
 
+    refobject_unref(source->identifier);
     free (source->mount);
     free (source);
 
@@ -323,7 +335,7 @@ void source_free_source (source_t *source)
 }
 
 
-client_t *source_find_client(source_t *source, int id)
+client_t *source_find_client(source_t *source, connection_id_t id)
 {
     client_t fakeclient;
     void *result;
@@ -343,120 +355,124 @@ client_t *source_find_client(source_t *source, int id)
     return NULL;
 }
 
+static inline int source_move_clients__single(source_t *source, source_t *dest, avl_tree *from, avl_tree *to, client_t *client, navigation_direction_t direction) {
+    if (navigation_history_navigate_to(&(client->history), dest->identifier, direction) != 0) {
+        ICECAST_LOG_DWARN("Can not change history: navigation of client=%p{.con->id=%llu, ...} from source=%p{.mount=%#H, ...} to dest=%p{.mount=%#H, ...} with direction %s failed",
+                client, (unsigned long long int)client->con->id, source, source->mount, dest, dest->mount, navigation_direction_to_str(direction));
+        return -1;
+    }
+
+    avl_delete(from, client, NULL);
+
+    /* when switching a client to a different queue, be wary of the
+     * refbuf it's referring to, if it's http headers then we need
+     * to write them so don't release it.
+     */
+    if (client->check_buffer != format_check_http_buffer) {
+        client_set_queue(client, NULL);
+        client->check_buffer = format_check_file_buffer;
+        if (source->con == NULL)
+            client->intro_offset = -1;
+    }
+
+    avl_insert(to, (void *)client);
+    return 0;
+}
 
 /* Move clients from source to dest provided dest is running
  * and that the stream format is the same.
  * The only lock that should be held when this is called is the
  * source tree lock
  */
-void source_move_clients (source_t *source, source_t *dest)
+void source_move_clients(source_t *source, source_t *dest, connection_id_t *id, navigation_direction_t direction)
 {
     unsigned long count = 0;
-    if (strcmp (source->mount, dest->mount) == 0)
-    {
+    if (strcmp(source->mount, dest->mount) == 0) {
         ICECAST_LOG_WARN("src and dst are the same \"%s\", skipping", source->mount);
         return;
     }
     /* we don't want the two write locks to deadlock in here */
-    thread_mutex_lock (&move_clients_mutex);
+    thread_mutex_lock(&move_clients_mutex);
 
     /* if the destination is not running then we can't move clients */
 
-    avl_tree_wlock (dest->pending_tree);
-    if (dest->running == 0 && dest->on_demand == 0)
-    {
+    avl_tree_wlock(dest->pending_tree);
+    if (dest->running == 0 && dest->on_demand == 0) {
         ICECAST_LOG_WARN("destination mount %s not running, unable to move clients ", dest->mount);
-        avl_tree_unlock (dest->pending_tree);
-        thread_mutex_unlock (&move_clients_mutex);
+        avl_tree_unlock(dest->pending_tree);
+        thread_mutex_unlock(&move_clients_mutex);
         return;
     }
 
-    do
-    {
-        client_t *client;
+    /* we need to move the client and pending trees - we must take the
+     * locks in this order to avoid deadlocks */
+    avl_tree_wlock(source->pending_tree);
+    avl_tree_wlock(source->client_tree);
 
-        /* we need to move the client and pending trees - we must take the
-         * locks in this order to avoid deadlocks */
-        avl_tree_wlock (source->pending_tree);
-        avl_tree_wlock (source->client_tree);
-
-        if (source->on_demand == 0 && source->format == NULL)
-        {
+    do {
+        if (source->on_demand == 0 && source->format == NULL) {
             ICECAST_LOG_INFO("source mount %s is not available", source->mount);
             break;
         }
-        if (source->format && dest->format)
-        {
-            if (source->format->type != dest->format->type)
-            {
+
+        if (source->format && dest->format) {
+            if (source->format->type != dest->format->type) {
                 ICECAST_LOG_WARN("stream %s and %s are of different types, ignored", source->mount, dest->mount);
                 break;
             }
         }
 
-        while (1)
-        {
-            avl_node *node = avl_get_first (source->pending_tree);
-            if (node == NULL)
-                break;
-            client = (client_t *)(node->key);
-            avl_delete (source->pending_tree, client, NULL);
+        if (id) {
+            client_t fakeclient;
+            connection_t fakecon;
+            void *result;
 
-            /* when switching a client to a different queue, be wary of the 
-             * refbuf it's referring to, if it's http headers then we need
-             * to write them so don't release it.
-             */
-            if (client->check_buffer != format_check_http_buffer)
-            {
-                client_set_queue (client, NULL);
-                client->check_buffer = format_check_file_buffer;
-                if (source->con == NULL)
-                    client->intro_offset = -1;
+            fakeclient.con = &fakecon;
+            fakeclient.con->id = *id;
+
+            if (avl_get_by_key(source->client_tree, &fakeclient, &result) == 0) {
+                if (source_move_clients__single(source, dest, source->client_tree, dest->pending_tree, result, direction) == 0)
+                    count++;
+            }
+        } else {
+            avl_node *next;
+
+            next = avl_get_first(source->pending_tree);
+            while (next) {
+                avl_node *node = next;
+
+                next = avl_get_next(next);
+
+                if (source_move_clients__single(source, dest, source->pending_tree, dest->pending_tree, node->key, direction) == 0)
+                    count++;
             }
 
-            avl_insert (dest->pending_tree, (void *)client);
-            count++;
-        }
+            next = avl_get_first(source->client_tree);
+            while (next) {
+                avl_node *node = next;
 
-        while (1)
-        {
-            avl_node *node = avl_get_first (source->client_tree);
-            if (node == NULL)
-                break;
+                next = avl_get_next(next);
 
-            client = (client_t *)(node->key);
-            avl_delete (source->client_tree, client, NULL);
-
-            /* when switching a client to a different queue, be wary of the 
-             * refbuf it's referring to, if it's http headers then we need
-             * to write them so don't release it.
-             */
-            if (client->check_buffer != format_check_http_buffer)
-            {
-                client_set_queue (client, NULL);
-                client->check_buffer = format_check_file_buffer;
-                if (source->con == NULL)
-                    client->intro_offset = -1;
+                if (source_move_clients__single(source, dest, source->client_tree, dest->pending_tree, node->key, direction) == 0)
+                    count++;
             }
-            avl_insert (dest->pending_tree, (void *)client);
-            count++;
         }
+
         ICECAST_LOG_INFO("passing %lu listeners to \"%s\"", count, dest->mount);
 
-        source->listeners = 0;
-        stats_event (source->mount, "listeners", "0");
-
+        source->listeners -= count;
+        stats_event_sub(source->mount, "listeners", count);
     } while (0);
 
-    avl_tree_unlock (source->pending_tree);
-    avl_tree_unlock (source->client_tree);
+    avl_tree_unlock(source->pending_tree);
+    avl_tree_unlock(source->client_tree);
 
     /* see if we need to wake up an on-demand relay */
     if (dest->running == 0 && dest->on_demand && count)
         dest->on_demand_req = 1;
 
-    avl_tree_unlock (dest->pending_tree);
-    thread_mutex_unlock (&move_clients_mutex);
+    avl_tree_unlock(dest->pending_tree);
+    thread_mutex_unlock(&move_clients_mutex);
 }
 
 
@@ -516,8 +532,7 @@ static refbuf_t *get_next_buffer (source_t *source)
         }
         source->last_read = current;
         refbuf = source->format->get_buffer (source);
-        if (source->client->con && source->client->con->error)
-        {
+        if (client_body_eof(source->client)) {
             ICECAST_LOG_INFO("End of Stream %s", source->mount);
             source->running = 0;
             continue;
@@ -533,8 +548,8 @@ static refbuf_t *get_next_buffer (source_t *source)
 /* general send routine per listener.  The deletion_expected tells us whether
  * the last in the queue is about to disappear, so if this client is still
  * referring to it after writing then drop the client as it's fallen too far
- * behind 
- */ 
+ * behind
+ */
 static void send_to_listener (source_t *source, client_t *client, int deletion_expected)
 {
     int bytes;
@@ -566,12 +581,12 @@ static void send_to_listener (source_t *source, client_t *client, int deletion_e
 
         loop--;
 
-        if (client->check_buffer (source, client) < 0)
+        if (client->check_buffer(source, client) < 0)
             break;
 
-        bytes = client->write_to_client (client);
+        bytes = client->write_to_client(client);
         if (bytes <= 0)
-            break;  /* can't write any more */
+            break; /* can't write any more */
 
         total_written += bytes;
     }
@@ -617,21 +632,8 @@ static FILE * source_open_dumpfile(const char * filename) {
  */
 static void source_init (source_t *source)
 {
-    ice_config_t *config = config_get_config();
-    char *listenurl;
+    char listenurl[512];
     const char *str;
-    int listen_url_size;
-    mount_proxy *mountinfo;
-
-    /* 6 for max size of port */
-    listen_url_size = strlen("http://") + strlen(config->hostname) +
-        strlen(":") + 6 + strlen(source->mount) + 1;
-
-    listenurl = malloc (listen_url_size);
-    memset (listenurl, '\000', listen_url_size);
-    snprintf (listenurl, listen_url_size, "http://%s:%d%s",
-            config->hostname, config->port, source->mount);
-    config_release_config();
 
     str = httpp_getvar(source->parser, "ice-audio-info");
     source->audio_info = util_dict_new();
@@ -641,10 +643,9 @@ static void source_init (source_t *source)
         stats_event (source->mount, "audio_info", str);
     }
 
+    client_get_baseurl(NULL, NULL, listenurl, sizeof(listenurl), NULL, NULL, NULL, NULL, NULL);
     stats_event (source->mount, "listenurl", listenurl);
 
-    free(listenurl);
-
     if (source->dumpfilename != NULL)
     {
         source->dumpfile = source_open_dumpfile (source->dumpfilename);
@@ -672,14 +673,7 @@ static void source_init (source_t *source)
     source->prev_listeners = -1;
     source->running = 1;
 
-    mountinfo = config_find_mount (config_get_config(), source->mount, MOUNT_TYPE_NORMAL);
-    if (mountinfo)
-    {
-        if (mountinfo->on_connect)
-            source_run_script (mountinfo->on_connect, source->mount);
-        auth_stream_start (mountinfo, source->mount);
-    }
-    config_release_config();
+    event_emit_clientevent("source-connect", source->client, source->mount);
 
     /*
     ** Now, if we have a fallback source and override is on, we want
@@ -688,15 +682,27 @@ static void source_init (source_t *source)
     ** loop or jingle track or whatever the fallback is used for
     */
 
-    if (source->fallback_override && source->fallback_mount)
-    {
+    ICECAST_LOG_DDEBUG("source=%p{.mount=%#H, .fallback_override=%i, .fallback_mount=%#H, ...}", source, source->mount, (int)source->fallback_override, source->fallback_mount);
+    if (source->fallback_override != FALLBACK_OVERRIDE_NONE && source->fallback_mount) {
         source_t *fallback_source;
 
         avl_tree_rlock(global.source_tree);
         fallback_source = source_find_mount(source->fallback_mount);
 
-        if (fallback_source)
-            source_move_clients (fallback_source, source);
+        if (fallback_source) {
+            ICECAST_LOG_DDEBUG("source=%p{.mount=%#H, .fallback_override=%i, ...}, fallback_source=%p{.mount=%#H, ...}", source, source->mount, (int)source->fallback_override, fallback_source, fallback_source->mount);
+            switch (source->fallback_override) {
+                case FALLBACK_OVERRIDE_NONE:
+                    /* no-op */
+                break;
+                case FALLBACK_OVERRIDE_ALL:
+                    source_move_clients(fallback_source, source, NULL, NAVIGATION_DIRECTION_REPLACE_CURRENT);
+                break;
+                case FALLBACK_OVERRIDE_OWN:
+                    source_move_clients(fallback_source, source, NULL, NAVIGATION_DIRECTION_UP);
+                break;
+            }
+        }
 
         avl_tree_unlock(global.source_tree);
     }
@@ -732,7 +738,7 @@ void source_main (source_t *source)
             source->stream_data_tail = refbuf;
             source->queue_size += refbuf->len;
             /* new buffer is referenced for burst */
-            refbuf_addref (refbuf);
+            refbuf_addref(refbuf);
 
             /* new data on queue, so check the burst point */
             source->burst_offset += refbuf->len;
@@ -744,7 +750,7 @@ void source_main (source_t *source)
                 {
                     source->burst_point = to_release->next;
                     source->burst_offset -= to_release->len;
-                    refbuf_release (to_release);
+                    refbuf_release(to_release);
                     continue;
                 }
                 break;
@@ -752,7 +758,7 @@ void source_main (source_t *source)
 
             /* save stream to file */
             if (source->dumpfile && source->format->write_buf_to_file)
-                source->format->write_buf_to_file (source, refbuf);
+                source->format->write_buf_to_file(source, refbuf);
         }
         /* lets see if we have too much data in the queue, but don't remove it until later */
         thread_mutex_lock(&source->lock);
@@ -768,15 +774,15 @@ void source_main (source_t *source)
 
         client_node = avl_get_first(source->client_tree);
         while (client_node) {
-            client = (client_t *)client_node->key;
+            client = (client_t *) client_node->key;
 
-            send_to_listener (source, client, remove_from_q);
+            send_to_listener(source, client, remove_from_q);
 
             if (client->con->error) {
                 client_node = avl_get_next(client_node);
                 if (client->respcode == 200)
-                    stats_event_dec (NULL, "listeners");
-                avl_delete(source->client_tree, (void *)client, _free_client);
+                    stats_event_dec(NULL, "listeners");
+                avl_delete(source->client_tree, (void *) client, _free_client);
                 source->listeners--;
                 ICECAST_LOG_DEBUG("Client removed");
                 continue;
@@ -788,8 +794,8 @@ void source_main (source_t *source)
         client_node = avl_get_first(source->pending_tree);
         while (client_node) {
 
-            if(source->max_listeners != -1 && 
-                    source->listeners >= (unsigned long)source->max_listeners) 
+            if(source->max_listeners != -1 &&
+                    source->listeners >= (unsigned long)source->max_listeners)
             {
                 /* The common case is caught in the main connection handler,
                  * this deals with rarer cases (mostly concerning fallbacks)
@@ -804,7 +810,7 @@ void source_main (source_t *source)
                         "mountpoint (%s).", source->mount);
                 continue;
             }
-            
+
             /* Otherwise, the client is accepted, add it */
             avl_insert(source->client_tree, client_node->key);
 
@@ -817,8 +823,8 @@ void source_main (source_t *source)
 
         /** clear pending tree **/
         while (avl_get_first(source->pending_tree)) {
-            avl_delete(source->pending_tree, 
-                    avl_get_first(source->pending_tree)->key, 
+            avl_delete(source->pending_tree,
+                    avl_get_first(source->pending_tree)->key,
                     source_remove_client);
         }
 
@@ -875,8 +881,6 @@ void source_main (source_t *source)
 
 static void source_shutdown (source_t *source)
 {
-    mount_proxy *mountinfo;
-
     source->running = 0;
     if (source->con && source->con->ip) {
         ICECAST_LOG_INFO("Source from %s at \"%s\" exiting", source->con->ip, source->mount);
@@ -884,14 +888,7 @@ static void source_shutdown (source_t *source)
         ICECAST_LOG_INFO("Source at \"%s\" exiting", source->mount);
     }
 
-    mountinfo = config_find_mount (config_get_config(), source->mount, MOUNT_TYPE_NORMAL);
-    if (mountinfo)
-    {
-        if (mountinfo->on_disconnect)
-            source_run_script (mountinfo->on_disconnect, source->mount);
-        auth_stream_end (mountinfo, source->mount);
-    }
-    config_release_config();
+    event_emit_clientevent("source-disconnect", source->client, source->mount);
 
     /* we have de-activated the source now, so no more clients will be
      * added, now move the listeners we have to the fallback (if any)
@@ -901,24 +898,49 @@ static void source_shutdown (source_t *source)
         source_t *fallback_source;
 
         avl_tree_rlock(global.source_tree);
-        fallback_source = source_find_mount (source->fallback_mount);
+        fallback_source = source_find_mount(source->fallback_mount);
 
         if (fallback_source != NULL)
-            source_move_clients (source, fallback_source);
+            source_move_clients(source, fallback_source, NULL, NAVIGATION_DIRECTION_DOWN);
 
-        avl_tree_unlock (global.source_tree);
+        avl_tree_unlock(global.source_tree);
     }
 
     /* delete this sources stats */
     stats_event(source->mount, NULL, NULL);
 
+    if (source->client && source->parser) {
+        /* For PUT support we check for 100-continue and send back a final 200. */
+        const char *expectcontinue = httpp_getvar(source->parser, "expect");
+
+        if (expectcontinue != NULL) {
+#ifdef HAVE_STRCASESTR
+            if (strcasestr (expectcontinue, "100-continue") != NULL)
+#else
+            ICECAST_LOG_WARN("OS doesn't support case insensitive substring checks...");
+            if (strstr (expectcontinue, "100-continue") != NULL)
+#endif
+            {
+                client_t *client = source->client;
+                source->client = NULL; /* detach client from source. */
+
+                util_http_build_header(client->refbuf->data, PER_CLIENT_REFBUF_SIZE, 0, 0, 200, NULL, NULL, NULL, "", NULL, source->client);
+                client->refbuf->len = strlen(client->refbuf->data);
+                refbuf_release(client->refbuf->next);
+                client->refbuf->next = NULL;
+                client->pos = 0;
+                fserve_add_client(client, NULL);
+            }
+        }
+    }
+
     /* we don't remove the source from the tree here, it may be a relay and
-       therefore reserved */
-    source_clear_source (source);
+     therefore reserved */
+    source_clear_source(source);
 
     global_lock();
     global.sources--;
-    stats_event_args (NULL, "sources", "%d", global.sources);
+    stats_event_args(NULL, "sources", "%d", global.sources);
     global_unlock();
 
     /* release our hold on the lock so the main thread can continue cleaning up */
@@ -926,20 +948,6 @@ static void source_shutdown (source_t *source)
 }
 
 
-static int _compare_clients(void *compare_arg, void *a, void *b)
-{
-    client_t *clienta = (client_t *)a;
-    client_t *clientb = (client_t *)b;
-
-    connection_t *cona = clienta->con;
-    connection_t *conb = clientb->con;
-
-    if (cona->id < conb->id) return -1;
-    if (cona->id > conb->id) return 1;
-
-    return 0;
-}
-
 int source_remove_client(void *key)
 {
     return 1;
@@ -949,12 +957,18 @@ static int _free_client(void *key)
 {
     client_t *client = (client_t *)key;
 
-    /* if no response has been sent then send a 404 */
-    if (client->respcode == 0)
-        client_send_404 (client, "Mount unavailable");
-    else
-        client_destroy(client);
-    
+    switch (client->respcode) {
+        case 0:
+            /* if no response has been sent then send a 404 */
+            client_send_error_by_id(client, ICECAST_ERROR_SOURCE_MOUNT_UNAVAILABLE);
+            break;
+        case 500:
+            client_send_error_by_id(client, ICECAST_ERROR_SOURCE_STREAM_PREPARATION_ERROR);
+            break;
+        default:
+            client_destroy(client);
+            break;
+    }
     return 1;
 }
 
@@ -982,7 +996,6 @@ static void _parse_audio_info (source_t *source, const char *s)
             if (esc)
             {
                 util_dict_set (source->audio_info, name, esc);
-                stats_event (source->mount, name, esc);
                 free (esc);
             }
         }
@@ -992,11 +1005,12 @@ static void _parse_audio_info (source_t *source, const char *s)
 
 
 /* Apply the mountinfo details to the source */
-static void source_apply_mount (source_t *source, mount_proxy *mountinfo)
+static void source_apply_mount (ice_config_t *config, source_t *source, mount_proxy *mountinfo)
 {
     const char *str;
     int val;
     http_parser_t *parser = NULL;
+    acl_t *acl = NULL;
 
     ICECAST_LOG_DEBUG("Applying mount information for \"%s\"", source->mount);
     avl_tree_rlock (source->client_tree);
@@ -1007,8 +1021,11 @@ static void source_apply_mount (source_t *source, mount_proxy *mountinfo)
         source->max_listeners = mountinfo->max_listeners;
         source->fallback_override = mountinfo->fallback_override;
         source->hidden = mountinfo->hidden;
+        source->allow_direct_access = mountinfo->allow_direct_access;
     }
 
+    stats_event(source->mount, "allow-direct-access", source->allow_direct_access ? "true" : "false");
+
     /* if a setting is available in the mount details then use it, else
      * check the parser details. */
 
@@ -1145,10 +1162,15 @@ static void source_apply_mount (source_t *source, mount_proxy *mountinfo)
     if (mountinfo && mountinfo->subtype)
         stats_event (source->mount, "subtype", mountinfo->subtype);
 
-    if (mountinfo && mountinfo->auth)
-        stats_event (source->mount, "authenticator", mountinfo->auth->type);
+    if (mountinfo)
+        acl = auth_stack_get_anonymous_acl(mountinfo->authstack, httpp_req_get);
+    if (!acl)
+        acl = auth_stack_get_anonymous_acl(config->authstack, httpp_req_get);
+    if (acl && acl_test_web(acl) == ACL_POLICY_DENY)
+        stats_event (source->mount, "authenticator", "(dummy)");
     else
         stats_event (source->mount, "authenticator", NULL);
+    acl_release(acl);
 
     if (mountinfo && mountinfo->fallback_mount)
     {
@@ -1201,12 +1223,15 @@ static void source_apply_mount (source_t *source, mount_proxy *mountinfo)
         source->timeout = mountinfo->source_timeout;
 
     if (mountinfo && mountinfo->burst_size >= 0)
-        source->burst_size = (unsigned int)mountinfo->burst_size;
+        source->burst_size = (unsigned int) mountinfo->burst_size;
 
     if (mountinfo && mountinfo->fallback_when_full)
         source->fallback_when_full = mountinfo->fallback_when_full;
 
-    avl_tree_unlock (source->client_tree);
+    if (mountinfo && mountinfo->max_history > 0)
+        playlist_set_max_tracks(source->history, mountinfo->max_history);
+
+    avl_tree_unlock(source->client_tree);
 }
 
 
@@ -1232,7 +1257,7 @@ void source_update_settings (ice_config_t *config, source_t *source, mount_proxy
     stats_event_args (source->mount, "listenurl", "http://%s:%d%s",
             config->hostname, config->port, source->mount);
 
-    source_apply_mount (source, mountinfo);
+    source_apply_mount (config, source, mountinfo);
 
     if (source->fallback_mount)
         ICECAST_LOG_DEBUG("fallback %s", source->fallback_mount);
@@ -1240,10 +1265,6 @@ void source_update_settings (ice_config_t *config, source_t *source, mount_proxy
         ICECAST_LOG_DEBUG("intro file is %s", mountinfo->intro_filename);
     if (source->dumpfilename)
         ICECAST_LOG_DEBUG("Dumping stream to %s", source->dumpfilename);
-    if (mountinfo && mountinfo->on_connect)
-        ICECAST_LOG_DEBUG("connect script \"%s\"", mountinfo->on_connect);
-    if (mountinfo && mountinfo->on_disconnect)
-        ICECAST_LOG_DEBUG("disconnect script \"%s\"", mountinfo->on_disconnect);
     if (source->on_demand)
     {
         ICECAST_LOG_DEBUG("on_demand set");
@@ -1299,7 +1320,6 @@ void source_client_callback (client_t *client, void *arg)
 {
     const char *agent;
     source_t *source = arg;
-    refbuf_t *old_data = client->refbuf;
 
     if (client->con->error)
     {
@@ -1310,9 +1330,9 @@ void source_client_callback (client_t *client, void *arg)
         source_free_source (source);
         return;
     }
-    client->refbuf = old_data->associated;
-    old_data->associated = NULL;
-    refbuf_release (old_data);
+
+    client->refbuf->len = 0;
+
     stats_event (source->mount, "source_ip", source->client->con->ip);
     agent = httpp_getvar (source->client->parser, "user-agent");
     if (agent)
@@ -1322,74 +1342,6 @@ void source_client_callback (client_t *client, void *arg)
             source, THREAD_DETACHED);
 }
 
-
-#ifndef _WIN32
-/* this sets up the new environment for script execution.
- * We ignore most failtures as we can not handle them anyway.
- */
-static inline void __setup_empty_script_environment(void) {
-    int i;
-
-    /* close at least the first 1024 handles */
-    for (i = 0; i < 1024; i++)
-        close(i);
-
-    /* open null device */
-    i = open("/dev/null", O_RDWR);
-    if (i != -1) {
-        /* attach null device to stdin, stdout and stderr */
-        if (i != 0)
-            dup2(i, 0);
-        if (i != 1)
-            dup2(i, 1);
-        if (i != 2)
-            dup2(i, 2);
-
-        /* close null device */
-        if (i > 2)
-            close(i);
-    }
-}
-
-static void source_run_script (char *command, char *mountpoint)
-{
-    pid_t pid, external_pid;
-
-    /* do a fork twice so that the command has init as parent */
-    external_pid = fork();
-    switch (external_pid)
-    {
-        case 0:
-            switch (pid = fork ())
-            {
-                case -1:
-                    ICECAST_LOG_ERROR("Unable to fork %s (%s)", command, strerror (errno));
-                    break;
-                case 0:  /* child */
-                    if (access(command, R_OK|X_OK) != 0) {
-                        ICECAST_LOG_ERROR("Unable to run command %s (%s)", command, strerror(errno));
-                        exit(1);
-                    }
-                    ICECAST_LOG_DEBUG("Starting command %s", command);
-                    __setup_empty_script_environment();
-                    /* consider to add action here as well */
-                    execl(command, command, mountpoint, (char *)NULL);
-                    exit(1);
-                default: /* parent */
-                    break;
-            }
-            exit (0);
-        case -1:
-            ICECAST_LOG_ERROR("Unable to fork %s", strerror (errno));
-            break;
-        default: /* parent */
-            waitpid (external_pid, NULL, 0);
-            break;
-    }
-}
-#endif
-
-
 static void *source_fallback_file (void *arg)
 {
     char *mount = arg;
@@ -1472,7 +1424,7 @@ void source_recheck_mounts (int update_all)
     for (; mount; mount = mount->next)
     {
         if (mount->mounttype != MOUNT_TYPE_NORMAL)
-	    continue;
+            continue;
 
         source_t *source = source_find_mount (mount->mountname);
 
@@ -1513,4 +1465,3 @@ void source_recheck_mounts (int update_all)
     avl_tree_unlock (global.source_tree);
     config_release_config();
 }
-
