@@ -3,12 +3,12 @@
  * This program is distributed under the GNU General Public License, version 2.
  * A copy of this license is included with this source.
  *
- * Copyright 2000-2004, Jack Moffitt <jack@xiph.org, 
+ * Copyright 2000-2004, Jack Moffitt <jack@xiph.org,
  *                      Michael Smith <msmith@xiph.org>,
  *                      oddsock <oddsock@xiph.org>,
  *                      Karl Heyes <karl@xiph.org>
  *                      and others (see AUTHORS for details).
- * Copyright 2012-2014, Philipp "ph3-der-loewe" Schafft <lion@lion.leolix.org>,
+ * Copyright 2012-2022, Philipp "ph3-der-loewe" Schafft <lion@lion.leolix.org>,
  */
 
 #ifdef HAVE_CONFIG_H
@@ -19,25 +19,27 @@
 #include <string.h>
 #include <stdlib.h>
 #include <stdarg.h>
+#include <ctype.h>
 
 #include <libxml/xmlmemory.h>
 #include <libxml/parser.h>
 #include <libxml/tree.h>
 
-#include "thread/thread.h"
-#include "avl/avl.h"
-#include "httpp/httpp.h"
-#include "net/sock.h"
+#include "common/thread/thread.h"
+#include "common/avl/avl.h"
+#include "common/httpp/httpp.h"
+#include "common/net/sock.h"
 
+#include "stats.h"
 #include "connection.h"
-
 #include "source.h"
 #include "global.h"
 #include "refbuf.h"
 #include "client.h"
-#include "stats.h"
+#include "admin.h"
 #include "xslt.h"
 #include "util.h"
+#include "auth.h"
 #define CATMODULE "stats"
 #include "logging.h"
 
@@ -85,15 +87,16 @@ static volatile event_listener_t *_event_listeners;
 
 
 static void *_stats_thread(void *arg);
-static int _compare_stats(void *a, void *b, void *arg);
-static int _compare_source_stats(void *a, void *b, void *arg);
+static int _compare_stats(void *arg, void *a, void *b);
+static int _compare_source_stats(void *arg, void *a, void *b);
 static int _free_stats(void *key);
 static int _free_source_stats(void *key);
 static void _add_event_to_queue(stats_event_t *event, event_queue_t *queue);
 static stats_node_t *_find_node(avl_tree *tree, const char *name);
 static stats_source_t *_find_source(avl_tree *tree, const char *source);
 static void _free_event(stats_event_t *event);
-static stats_event_t *_get_event_from_queue (event_queue_t *queue);
+static stats_event_t *_get_event_from_queue(event_queue_t *queue);
+static void __add_metadata(xmlNodePtr node, const char *tag);
 
 
 /* simple helper function for creating an event */
@@ -135,7 +138,7 @@ void stats_initialize(void)
     thread_mutex_create(&_stats_mutex);
 
     /* set up stats queues */
-    event_queue_init (&_global_event_queue);
+    event_queue_init(&_global_event_queue);
     thread_mutex_create(&_global_event_mutex);
 
     /* fire off the stats thread */
@@ -147,11 +150,13 @@ void stats_shutdown(void)
 {
     int n;
 
-    if(!_stats_running) /* We can't shutdown if we're not running. */
+    if (!_stats_running) /* We can't shutdown if we're not running. */
         return;
 
     /* wait for thread to exit */
+    thread_mutex_lock(&_stats_mutex);
     _stats_running = 0;
+    thread_mutex_unlock(&_stats_mutex);
     thread_join(_stats_thread_id);
 
     /* wait for other threads to shut down */
@@ -189,7 +194,7 @@ void stats_shutdown(void)
 stats_t *stats_get_stats(void)
 {
     /* lock global stats
-    
+
      copy stats
 
      unlock global stats
@@ -209,9 +214,9 @@ void stats_event(const char *source, const char *name, const char *value)
         ICECAST_LOG_WARN("seen non-UTF8 data, probably incorrect metadata (%s, %s)", name, value);
         return;
     }
-    event = build_event (source, name, value);
+    event = build_event(source, name, value);
     if (event)
-        queue_global_event (event);
+        queue_global_event(event);
 }
 
 
@@ -329,7 +334,7 @@ void stats_event_add(const char *source, const char *name, unsigned long value)
     if (event)
     {
         event->value = malloc (16);
-        snprintf (event->value, 16, "%ld", value);
+        snprintf(event->value, 16, "%lu", value);
         event->action = STATS_EVENT_ADD;
         queue_global_event (event);
     }
@@ -341,7 +346,7 @@ void stats_event_sub(const char *source, const char *name, unsigned long value)
     if (event)
     {
         event->value = malloc (16);
-        snprintf (event->value, 16, "%ld", value);
+        snprintf(event->value, 16, "%lu", value);
         event->action = STATS_EVENT_SUB;
         queue_global_event (event);
     }
@@ -370,18 +375,18 @@ static stats_node_t *_find_node(avl_tree *stats_tree, const char *name)
 
     /* get the root node */
     node = stats_tree->root->right;
-    
+
     while (node) {
-        stats = (stats_node_t *)node->key;
+        stats = (stats_node_t *) node->key;
         cmp = strcmp(name, stats->name);
-        if (cmp < 0) 
+        if (cmp < 0)
             node = node->left;
         else if (cmp > 0)
             node = node->right;
         else
             return stats;
     }
-    
+
     /* didn't find it */
     return NULL;
 }
@@ -398,7 +403,7 @@ static stats_source_t *_find_source(avl_tree *source_tree, const char *source)
     /* get the root node */
     node = source_tree->root->right;
     while (node) {
-        stats = (stats_source_t *)node->key;
+        stats = (stats_source_t *) node->key;
         cmp = strcmp(source, stats->source);
         if (cmp < 0)
             node = node->left;
@@ -415,7 +420,7 @@ static stats_source_t *_find_source(avl_tree *source_tree, const char *source)
 static stats_event_t *_copy_event(stats_event_t *event)
 {
     stats_event_t *copy = (stats_event_t *)calloc(1, sizeof(stats_event_t));
-    if (event->source) 
+    if (event->source)
         copy->source = (char *)strdup(event->source);
     else
         copy->source = NULL;
@@ -433,7 +438,7 @@ static stats_event_t *_copy_event(stats_event_t *event)
 
 
 /* helper to apply specialised changes to a stats node */
-static void modify_node_event (stats_node_t *node, stats_event_t *event)
+static void modify_node_event(stats_node_t *node, stats_event_t *event)
 {
     char *str;
 
@@ -531,7 +536,7 @@ static void process_source_event (stats_event_t *event)
         else
             snode->hidden = 0;
 
-        avl_insert(_stats.source_tree, (void *)snode);
+        avl_insert(_stats.source_tree, (void *) snode);
     }
     if (event->name)
     {
@@ -636,7 +641,7 @@ static inline void __format_time(char * buffer, size_t len, const char * format)
 
 void stats_event_time (const char *mount, const char *name)
 {
-    char buffer[100];
+    char buffer[256];
 
     __format_time(buffer, sizeof(buffer), "%a, %d %b %Y %H:%M:%S ");
     stats_event (mount, name, buffer);
@@ -645,7 +650,7 @@ void stats_event_time (const char *mount, const char *name)
 
 void stats_event_time_iso8601 (const char *mount, const char *name)
 {
-    char buffer[100];
+    char buffer[256];
 
     __format_time(buffer, sizeof(buffer), "%Y-%m-%dT%H:%M:%S");
     stats_event (mount, name, buffer);
@@ -667,6 +672,8 @@ static void *_stats_thread(void *arg)
     stats_event_t *copy;
     event_listener_t *listener;
 
+    (void)arg;
+
     stats_event_time (NULL, "server_start");
     stats_event_time_iso8601 (NULL, "server_start_iso8601");
 
@@ -686,7 +693,14 @@ static void *_stats_thread(void *arg)
     stats_event (NULL, "listener_connections", "0");
 
     ICECAST_LOG_INFO("stats thread started");
-    while (_stats_running) {
+    while (1) {
+        thread_mutex_lock(&_stats_mutex);
+        if (!_stats_running) {
+            thread_mutex_unlock(&_stats_mutex);
+            break;
+        }
+        thread_mutex_unlock(&_stats_mutex);
+
         thread_mutex_lock(&_global_event_mutex);
         if (_global_event_queue.head != NULL) {
             /* grab the next event from the queue */
@@ -704,7 +718,7 @@ static void *_stats_thread(void *arg)
                 process_global_event (event);
             else
                 process_source_event (event);
-            
+
             /* now we have an event that's been processed into the running stats */
             /* this event should get copied to event listeners' queues */
             listener = (event_listener_t *)_event_listeners;
@@ -755,7 +769,7 @@ static void _unregister_listener(event_listener_t *listener)
 static stats_event_t *_make_event_from_node(stats_node_t *node, char *source)
 {
     stats_event_t *event = (stats_event_t *)malloc(sizeof(stats_event_t));
-    
+
     if (source != NULL)
         event->source = (char *)strdup(source);
     else
@@ -811,30 +825,76 @@ static int _send_event_to_client(stats_event_t *event, client_t *client)
     return 0;
 }
 
-
-static xmlNodePtr _dump_stats_to_doc (xmlNodePtr root, const char *show_mount, int hidden)
+void stats_add_authstack(auth_stack_t *stack, xmlNodePtr parent)
 {
+    xmlNodePtr authentication;
+    authentication = xmlNewTextChild(parent, NULL, XMLSTR("authentication"), NULL);
+
+    auth_stack_addref(stack);
+
+    while (stack) {
+        auth_t *auth = auth_stack_get(stack);
+        admin_add_role_to_authentication(auth, authentication);
+        auth_release(auth);
+        auth_stack_next(&stack);
+   }
+}
+
+static inline int __is_in_list(const char *key, const char *list[])
+{
+    size_t i;
+    for (i = 0; list[i]; i++)
+        if (strcmp(key, list[i]) == 0)
+            return 1;
+    return 0;
+}
+
+static inline int __include_node(unsigned int flags, const char *key, const char *list[])
+{
+    return !(flags & STATS_XML_FLAG_PUBLIC_VIEW) || __is_in_list(key, list);
+}
+
+static xmlNodePtr _dump_stats_to_doc (xmlNodePtr root, unsigned int flags, const char *show_mount, client_t *client) {
+    static const char *public_keys_global[] = {"admin", "location", "host", "server_id", "server_start_iso8601", NULL};
+    static const char *public_keys_source[] = {"listeners", "server_name", "server_description", "stream_start_iso8601", "subtype", "content-type", "listenurl", "genre", "display-title", NULL};
+    int hidden = flags & STATS_XML_FLAG_SHOW_HIDDEN ? 1 : 0;
     avl_node *avlnode;
     xmlNodePtr ret = NULL;
+    ice_config_t *config;
+
+    if (flags & STATS_XML_FLAG_PUBLIC_VIEW) {
+        /* Ensure those flags are clear when rendering a public view */
+        flags &= ~(STATS_XML_FLAG_SHOW_LISTENERS|STATS_XML_FLAG_SHOW_HIDDEN);
+    } else {
+        config = config_get_config();
+        stats_add_authstack(config->authstack, root);
+        config_release_config();
+    }
 
     thread_mutex_lock(&_stats_mutex);
     /* general stats first */
     avlnode = avl_get_first(_stats.global_tree);
-    while (avlnode)
-    {
+
+    while (avlnode) {
         stats_node_t *stat = avlnode->key;
-        if (stat->hidden <=  hidden)
+        if (stat->hidden <=  hidden && __include_node(flags, stat->name, public_keys_global))
             xmlNewTextChild (root, NULL, XMLSTR(stat->name), XMLSTR(stat->value));
         avlnode = avl_get_next (avlnode);
     }
     /* now per mount stats */
     avlnode = avl_get_first(_stats.source_tree);
-    while (avlnode)
-    {
+
+    while (avlnode) {
         stats_source_t *source = (stats_source_t *)avlnode->key;
+
         if (source->hidden <= hidden &&
                 (show_mount == NULL || strcmp (show_mount, source->source) == 0))
         {
+            xmlNodePtr metadata, history;
+            source_t *source_real;
+            mount_proxy *mountproxy;
+            int i;
+
             avl_node *avlnode2 = avl_get_first (source->stats_tree);
             xmlNodePtr xmlnode = xmlNewTextChild (root, NULL, XMLSTR("source"), NULL);
 
@@ -844,9 +904,47 @@ static xmlNodePtr _dump_stats_to_doc (xmlNodePtr root, const char *show_mount, i
             while (avlnode2)
             {
                 stats_node_t *stat = avlnode2->key;
-                xmlNewTextChild (xmlnode, NULL, XMLSTR(stat->name), XMLSTR(stat->value));
+                if (__include_node(flags, stat->name, public_keys_source)) {
+                    if (client && strcmp(stat->name, "listenurl") == 0) {
+                        char buf[512];
+                        client_get_baseurl(client, NULL, buf, sizeof(buf), NULL, NULL, NULL, source->source, NULL);
+                        xmlNewTextChild (xmlnode, NULL, XMLSTR(stat->name), XMLSTR(buf));
+                    } else {
+                        xmlNewTextChild (xmlnode, NULL, XMLSTR(stat->name), XMLSTR(stat->value));
+                    }
+                }
                 avlnode2 = avl_get_next (avlnode2);
             }
+
+
+            avl_tree_rlock(global.source_tree);
+            source_real = source_find_mount_raw(source->source);
+            if (source_real) {
+                history = playlist_render_xspf(source_real->history);
+                if (history)
+                    xmlAddChild(xmlnode, history);
+
+                metadata = xmlNewTextChild(xmlnode, NULL, XMLSTR("metadata"), NULL);
+                if (source_real->format) {
+                    for (i = 0; i < source_real->format->vc.comments; i++)
+                        __add_metadata(metadata, source_real->format->vc.user_comments[i]);
+                }
+
+                if (source_real->running)
+                    xmlNewTextChild(xmlnode, NULL, XMLSTR("content-type"), XMLSTR(source_real->format->contenttype));
+
+                if (flags & STATS_XML_FLAG_SHOW_LISTENERS)
+                    admin_add_listeners_to_mount(source_real, xmlnode, client->mode);
+            }
+            avl_tree_unlock(global.source_tree);
+
+            if (!(flags & STATS_XML_FLAG_PUBLIC_VIEW)) {
+                config = config_get_config();
+                mountproxy = config_find_mount(config, source->source, MOUNT_TYPE_NORMAL);
+                if (mountproxy)
+                    stats_add_authstack(mountproxy->authstack, xmlnode);
+                config_release_config();
+            }
         }
         avlnode = avl_get_next (avlnode);
     }
@@ -856,7 +954,7 @@ static xmlNodePtr _dump_stats_to_doc (xmlNodePtr root, const char *show_mount, i
 
 
 /* factoring out code for stats loops
-** this function copies all stats to queue, and registers 
+** this function copies all stats to queue, and registers
 ** the queue for all new events atomically.
 ** note: mutex must already be created!
 */
@@ -870,12 +968,12 @@ static void _register_listener (event_listener_t *listener)
     thread_mutex_lock(&_stats_mutex);
 
     /* first we fill our queue with the current stats */
-    
+
     /* start with the global stats */
     node = avl_get_first(_stats.global_tree);
     while (node) {
-        event = _make_event_from_node((stats_node_t *)node->key, NULL);
-        _add_event_to_queue (event, &listener->queue);
+        event = _make_event_from_node((stats_node_t *) node->key, NULL);
+        _add_event_to_queue(event, &listener->queue);
 
         node = avl_get_next(node);
     }
@@ -891,7 +989,7 @@ static void _register_listener (event_listener_t *listener)
 
             node2 = avl_get_next(node2);
         }
-        
+
         node = avl_get_next(node);
     }
 
@@ -921,7 +1019,14 @@ void *stats_connection(void *arg)
 
     _register_listener (&listener);
 
-    while (_stats_running) {
+    while (1) {
+        thread_mutex_lock(&_stats_mutex);
+        if (!_stats_running) {
+            thread_mutex_unlock(&_stats_mutex);
+            break;
+        }
+        thread_mutex_unlock(&_stats_mutex);
+
         thread_mutex_lock (&listener.mutex);
         event = _get_event_from_queue (&listener.queue);
         thread_mutex_unlock (&listener.mutex);
@@ -952,6 +1057,8 @@ void *stats_connection(void *arg)
 
 void stats_callback (client_t *client, void *notused)
 {
+    (void)notused;
+
     if (client->con->error)
     {
         client_destroy (client);
@@ -970,30 +1077,57 @@ typedef struct _source_xml_tag {
 } source_xml_t;
 
 
-void stats_transform_xslt(client_t *client, const char *uri)
+void stats_transform_xslt(client_t *client)
 {
     xmlDocPtr doc;
-    char *xslpath = util_get_path_from_normalised_uri (uri);
-    const char *mount = httpp_get_query_param (client->parser, "mount");
+    char *xslpath = util_get_path_from_normalised_uri(client->uri);
+    const char *mount = httpp_get_param(client->parser, "mount");
 
-    doc = stats_get_xml (0, mount);
+    doc = stats_get_xml(STATS_XML_FLAG_NONE, mount, client);
 
-    xslt_transform(doc, xslpath, client);
+    xslt_transform(doc, xslpath, client, 200, NULL, NULL);
 
     xmlFreeDoc(doc);
-    free (xslpath);
+    free(xslpath);
 }
 
-xmlDocPtr stats_get_xml(int show_hidden, const char *show_mount)
+static void __add_metadata(xmlNodePtr node, const char *tag) {
+    const char *value = strstr(tag, "=");
+    char *name = NULL;
+    size_t namelen = value - tag + 1;
+    size_t i;
+
+    if (!value)
+        return;
+
+    name = malloc(namelen);
+    if (!name)
+        return;
+
+    for (i = 0; i < (namelen - 1); i++)
+        name[i] = tolower(tag[i]);
+
+    name[namelen-1] = 0;
+
+    xmlNewTextChild(node, NULL, XMLSTR(name), XMLSTR(value+1));
+
+    free(name);
+}
+
+xmlDocPtr stats_get_xml(unsigned int flags, const char *show_mount, client_t *client)
 {
     xmlDocPtr doc;
     xmlNodePtr node;
+    xmlNodePtr modules;
 
     doc = xmlNewDoc (XMLSTR("1.0"));
     node = xmlNewDocNode (doc, NULL, XMLSTR("icestats"), NULL);
     xmlDocSetRootElement(doc, node);
 
-    node = _dump_stats_to_doc (node, show_mount, show_hidden);
+    modules = module_container_get_modulelist_as_xml(global.modulecontainer);
+    xmlAddChild(node, modules);
+
+    node = _dump_stats_to_doc(node, flags, show_mount, client);
 
     return doc;
 }
@@ -1004,6 +1138,8 @@ static int _compare_stats(void *arg, void *a, void *b)
     stats_node_t *nodea = (stats_node_t *)a;
     stats_node_t *nodeb = (stats_node_t *)b;
 
+    (void)arg;
+
     return strcmp(nodea->name, nodeb->name);
 }
 
@@ -1012,6 +1148,8 @@ static int _compare_source_stats(void *arg, void *a, void *b)
     stats_source_t *nodea = (stats_source_t *)a;
     stats_source_t *nodeb = (stats_source_t *)b;
 
+    (void)arg;
+
     return strcmp(nodea->source, nodeb->source);
 }
 
@@ -1021,7 +1159,7 @@ static int _free_stats(void *key)
     free(node->value);
     free(node->name);
     free(node);
-    
+
     return 1;
 }
 
@@ -1079,7 +1217,7 @@ refbuf_t *stats_get_streams (void)
         }
         node = avl_get_next(node);
     }
-    thread_mutex_unlock (&_stats_mutex);
+    thread_mutex_unlock(&_stats_mutex);
     cur->len = STREAMLIST_BLKSIZE - remaining;
     return start;
 }
