@@ -3,11 +3,12 @@
  * This program is distributed under the GNU General Public License, version 2.
  * A copy of this license is included with this source.
  *
- * Copyright 2000-2004, Jack Moffitt <jack@xiph.org, 
+ * Copyright 2000-2004, Jack Moffitt <jack@xiph.org,
  *                      Michael Smith <msmith@xiph.org>,
  *                      oddsock <oddsock@xiph.org>,
  *                      Karl Heyes <karl@xiph.org>
  *                      and others (see AUTHORS for details).
+ * Copyright 2013-2018, Philipp "ph3-der-loewe" Schafft <lion@lion.leolix.org>,
  */
 
 /* -*- c-basic-offset: 4; indent-tabs-mode: nil; -*- */
@@ -18,24 +19,24 @@
 #include <stdio.h>
 #include <string.h>
 #include <stdlib.h>
-#include <curl/curl.h>
 
-#include "thread/thread.h"
+#include "common/thread/thread.h"
 
-#include "connection.h"
-#include "refbuf.h"
-#include "client.h"
+#include "yp.h"
+#include "global.h"
+#include "curl.h"
 #include "logging.h"
-#include "format.h"
 #include "source.h"
 #include "cfgfile.h"
 #include "stats.h"
+#include "listensocket.h"
+#include "prng.h"
 
 #ifdef WIN32
 #define snprintf _snprintf
 #endif
 
-#define CATMODULE "yp" 
+#define CATMODULE "yp"
 
 struct yp_server
 {
@@ -44,6 +45,7 @@ struct yp_server
     unsigned    url_timeout;
     unsigned    touch_interval;
     int         remove;
+    char        *listen_socket_id;
 
     CURL *curl;
     struct ypdata_tag *mounts, *pending_mounts;
@@ -74,10 +76,10 @@ typedef struct ypdata_tag
     char *subtype;
 
     struct yp_server *server;
-    time_t      next_update;
-    unsigned    touch_interval;
-    char        *error_msg;
-    int    (*process)(struct ypdata_tag *yp, char *s, unsigned len);
+    time_t next_update;
+    unsigned touch_interval;
+    char *error_msg;
+    int (*process)(struct ypdata_tag *yp, char *s, unsigned len);
 
     struct ypdata_tag *next;
 } ypdata_t;
@@ -95,18 +97,22 @@ static volatile unsigned client_limit = 0;
 static volatile char *server_version = NULL;
 
 static void *yp_update_thread(void *arg);
-static void add_yp_info (ypdata_t *yp, void *info, int type);
-static int do_yp_remove (ypdata_t *yp, char *s, unsigned len);
-static int do_yp_add (ypdata_t *yp, char *s, unsigned len);
-static int do_yp_touch (ypdata_t *yp, char *s, unsigned len);
+static void add_yp_info(ypdata_t *yp, void *info, int type);
+static int do_yp_remove(ypdata_t *yp, char *s, unsigned len);
+static int do_yp_add(ypdata_t *yp, char *s, unsigned len);
+static int do_yp_touch(ypdata_t *yp, char *s, unsigned len);
+static void add_pending_yp (struct yp_server *server);
+static void delete_marked_yp(struct yp_server *server);
 static void yp_destroy_ypdata(ypdata_t *ypdata);
 
 
 /* curl callback used to parse headers coming back from the YP server */
-static int handle_returned_header (void *ptr, size_t size, size_t nmemb, void *stream)
+static size_t handle_returned_header (void *ptr, size_t size, size_t nmemb, void *stream)
 {
     ypdata_t *yp = stream;
-    unsigned bytes = size * nmemb;
+    size_t bytes = size * nmemb;
+
+    prng_write(ptr, bytes);
 
     /* ICECAST_LOG_DEBUG("header from YP is \"%.*s\"", bytes, ptr); */
     if (strncasecmp (ptr, "YPResponse: 1", 13) == 0)
@@ -114,7 +120,7 @@ static int handle_returned_header (void *ptr, size_t size, size_t nmemb, void *s
 
     if (strncasecmp (ptr, "YPMessage: ", 11) == 0)
     {
-        unsigned len = bytes - 11;
+        size_t len = bytes - 11;
         free (yp->error_msg);
         yp->error_msg = calloc (1, len);
         if (yp->error_msg)
@@ -125,7 +131,7 @@ static int handle_returned_header (void *ptr, size_t size, size_t nmemb, void *s
     {
         if (strncasecmp (ptr, "SID: ", 5) == 0)
         {
-            unsigned len = bytes - 5;
+            size_t len = bytes - 5;
             free (yp->sid);
             yp->sid = calloc (1, len);
             if (yp->sid)
@@ -142,14 +148,7 @@ static int handle_returned_header (void *ptr, size_t size, size_t nmemb, void *s
         ICECAST_LOG_DEBUG("server touch interval is %u", secs);
         yp->touch_interval = secs;
     }
-    return (int)bytes;
-}
-
-
-/* capture returned data, but don't do anything with it, shouldn't be any */
-static int handle_returned_data (void *ptr, size_t size, size_t nmemb, void *stream)
-{
-    return (int)(size*nmemb);
+    return (size_t)bytes;
 }
 
 
@@ -178,13 +177,28 @@ static struct yp_server *find_yp_server (const char *url)
 
 static void destroy_yp_server (struct yp_server *server)
 {
+    ypdata_t *yp;
+
     if (server == NULL)
         return;
     ICECAST_LOG_DEBUG("Removing YP server entry for %s", server->url);
-    if (server->curl)
-        curl_easy_cleanup (server->curl);
-    if (server->mounts) ICECAST_LOG_WARN("active ypdata not freed up");
-    if (server->pending_mounts) ICECAST_LOG_WARN("pending ypdata not freed up");
+
+    /* delete yps:
+     * first move all pendings into main queue.
+     * then mark all main queue entries for deleting.
+     * then remove all marked entries.
+     */
+    add_pending_yp(server);
+    yp = server->mounts;
+    while (yp) {
+        yp->remove = 1;
+        yp = yp->next;
+    }
+    delete_marked_yp(server);
+
+    icecast_curl_free(server->curl);
+    if (server->mounts) ICECAST_LOG_WARN("active ypdata not freed");
+    if (server->pending_mounts) ICECAST_LOG_WARN("pending ypdata not freed");
     free (server->url);
     free (server->server_id);
     free (server);
@@ -208,7 +222,6 @@ static ypdata_t *find_yp_mount (ypdata_t *mounts, const char *mount)
 
 void yp_recheck_config (ice_config_t *config)
 {
-    int i;
     struct yp_server *server;
 
     ICECAST_LOG_DEBUG("Updating YP configuration");
@@ -223,11 +236,11 @@ void yp_recheck_config (ice_config_t *config)
     client_limit = config->client_limit;
     free ((char*)server_version);
     server_version = strdup (config->server_id);
-    /* for each yp url in config, check to see if one exists 
+    /* for each yp url in config, check to see if one exists
        if not, then add it. */
-    for (i=0 ; i < config->num_yp_directories; i++)
+    for (yp_directory_t *yp = config->yp_directories; yp; yp = yp->next)
     {
-        server = find_yp_server (config->yp_url[i]);
+        server = find_yp_server (yp->url);
         if (server == NULL)
         {
             server = calloc (1, sizeof (struct yp_server));
@@ -238,10 +251,11 @@ void yp_recheck_config (ice_config_t *config)
                 break;
             }
             server->server_id = strdup ((char *)server_version);
-            server->url = strdup (config->yp_url[i]);
-            server->url_timeout = config->yp_url_timeout[i];
-            server->touch_interval = config->yp_touch_interval[i];
-            server->curl = curl_easy_init();
+            server->url = strdup (yp->url);
+            server->url_timeout = yp->timeout;
+            server->touch_interval = yp->touch_interval;
+            server->listen_socket_id = yp->listen_socket_id;
+            server->curl = icecast_curl_new(server->url, &(server->curl_error[0]));
             if (server->curl == NULL)
             {
                 destroy_yp_server (server);
@@ -251,16 +265,7 @@ void yp_recheck_config (ice_config_t *config)
                 server->url_timeout = 6;
             if (server->touch_interval < 30)
                 server->touch_interval = 30;
-            curl_easy_setopt (server->curl, CURLOPT_USERAGENT, server->server_id);
-            curl_easy_setopt (server->curl, CURLOPT_URL, server->url);
             curl_easy_setopt (server->curl, CURLOPT_HEADERFUNCTION, handle_returned_header);
-            curl_easy_setopt (server->curl, CURLOPT_WRITEFUNCTION, handle_returned_data);
-            curl_easy_setopt (server->curl, CURLOPT_WRITEDATA, server->curl);
-            curl_easy_setopt (server->curl, CURLOPT_TIMEOUT, server->url_timeout);
-            curl_easy_setopt (server->curl, CURLOPT_NOSIGNAL, 1L);
-            curl_easy_setopt (server->curl, CURLOPT_FOLLOWLOCATION, 1L);
-            curl_easy_setopt (server->curl, CURLOPT_MAXREDIRS, 3L);
-            curl_easy_setopt (server->curl, CURLOPT_ERRORBUFFER, &(server->curl_error[0]));
             server->next = (struct yp_server *)pending_yps;
             pending_yps = server;
             ICECAST_LOG_INFO("Adding new YP server \"%s\" (timeout %ds, default interval %ds)",
@@ -272,7 +277,9 @@ void yp_recheck_config (ice_config_t *config)
         }
     }
     thread_rwlock_unlock (&yp_lock);
+    thread_rwlock_wlock(&yp_lock);
     yp_update = 1;
+    thread_rwlock_unlock(&yp_lock);
 }
 
 
@@ -334,7 +341,7 @@ static int send_to_yp (const char *cmd, ypdata_t *yp, char *post)
             ICECAST_LOG_INFO("YP %s on %s failed: %s", cmd, server->url, yp->error_msg);
         }
         yp->process = do_yp_add;
-        free (yp->sid);
+        free(yp->sid);
         yp->sid = NULL;
         return -1;
     }
@@ -371,6 +378,12 @@ static int do_yp_add (ypdata_t *yp, char *s, unsigned len)
 {
     int ret;
     char *value;
+    ice_config_t *config;
+    char *admin;
+
+    config = config_get_config();
+    admin = util_url_escape(config->admin);
+    config_release_config();
 
     value = stats_get_value (yp->mount, "server_type");
     add_yp_info (yp, value, YP_SERVER_TYPE);
@@ -388,7 +401,7 @@ static int do_yp_add (ypdata_t *yp, char *s, unsigned len)
     add_yp_info (yp, value, YP_SERVER_GENRE);
     free (value);
 
-    value = stats_get_value (yp->mount, "bitrate");
+    value = stats_get_value (yp->mount, "audio_bitrate");
     if (value == NULL)
         value = stats_get_value (yp->mount, "ice-bitrate");
     add_yp_info (yp, value, YP_BITRATE);
@@ -406,11 +419,14 @@ static int do_yp_add (ypdata_t *yp, char *s, unsigned len)
     add_yp_info (yp, value, YP_AUDIO_INFO);
     free (value);
 
-    ret = snprintf (s, len, "action=add&sn=%s&genre=%s&cpswd=%s&desc="
+    ret = snprintf (s, len, "action=add&admin=%s&sn=%s&genre=%s&cpswd=%s&desc="
                     "%s&url=%s&listenurl=%s&type=%s&stype=%s&b=%s&%s\r\n",
+                    admin,
                     yp->server_name, yp->server_genre, yp->cluster_password,
                     yp->server_desc, yp->url, yp->listen_url,
                     yp->server_type, yp->subtype, yp->bitrate, yp->audio_info);
+    free(admin);
+
     if (ret >= (signed)len)
         return ret+1;
     ret = send_to_yp ("add", yp, s);
@@ -435,7 +451,7 @@ static int do_yp_touch (ypdata_t *yp, char *s, unsigned len)
     if (artist || title)
     {
          char *song;
-         char *separator = " - ";
+         const char *separator = " - ";
          if (artist == NULL)
          {
              artist = strdup("");
@@ -491,7 +507,7 @@ static int do_yp_touch (ypdata_t *yp, char *s, unsigned len)
 
 
 
-static int process_ypdata (struct yp_server *server, ypdata_t *yp)
+static int process_ypdata(struct yp_server *server, ypdata_t *yp)
 {
     unsigned len = 1024;
     char *s = NULL, *tmp;
@@ -503,7 +519,7 @@ static int process_ypdata (struct yp_server *server, ypdata_t *yp)
     while (1)
     {
         int ret;
-        if ((tmp = realloc (s, len)) == NULL)
+        if ((tmp = realloc(s, len)) == NULL)
             return 0;
         s = tmp;
 
@@ -552,16 +568,19 @@ static void yp_process_server (struct yp_server *server)
 
 
 
-static ypdata_t *create_yp_entry (const char *mount)
+static ypdata_t *create_yp_entry (struct yp_server *server, const char *mount)
 {
     ypdata_t *yp;
-    char *s;
+
+    if (!server)
+        return NULL;
 
     yp = calloc (1, sizeof (ypdata_t));
     do
     {
+        listensocket_t *listen_socket = NULL;
         unsigned len = 512;
-        int ret;
+        ssize_t ret;
         char *url;
         mount_proxy *mountproxy = NULL;
         ice_config_t *config;
@@ -584,15 +603,33 @@ static ypdata_t *create_yp_entry (const char *mount)
         url = malloc (len);
         if (url == NULL)
             break;
-        config = config_get_config();
-        ret = snprintf (url, len, "http://%s:%d%s", config->hostname, config->port, mount);
-        if (ret >= (signed)len)
-        {
-            s = realloc (url, ++ret);
-            if (s) url = s;
-            snprintf (url, ret, "http://%s:%d%s", config->hostname, config->port, mount);
-        }
 
+        if (server->listen_socket_id) {
+            listen_socket = listensocket_container_get_by_id(global.listensockets,
+                                                             server->listen_socket_id);
+            if (!listen_socket)
+                ICECAST_LOG_ERROR("Failure to find listen socket with ID %#H, using default.",
+                    server->listen_socket_id);
+        }
+        ret = client_get_baseurl(NULL, listen_socket, url, len, NULL, NULL, NULL, mount, NULL);
+        if (ret >= len) {
+            // Buffer was too small, allocate a big enough one
+            char *s = realloc (url, ret + 1);
+            if (!s) {
+                refobject_unref(listen_socket);
+                free(url);
+                break;
+            }
+            url = s;
+
+            ret = client_get_baseurl(NULL, listen_socket, url, len, NULL, NULL, NULL, mount, NULL);
+        }
+        refobject_unref(listen_socket);
+
+        if (ret < 0 || ret >= len)
+            break;
+
+        config = config_get_config();
         mountproxy = config_find_mount (config, mount, MOUNT_TYPE_NORMAL);
         if (mountproxy && mountproxy->cluster_password)
             add_yp_info (yp, mountproxy->cluster_password, YP_CLUSTER_PASSWORD);
@@ -622,10 +659,10 @@ static void check_servers (void)
         if (server->remove)
         {
             struct yp_server *to_go = server;
-            ICECAST_LOG_DEBUG("YP server \"%s\"removed", server->url);
+            ICECAST_LOG_DEBUG("YP server \"%s\" removed", server->url);
             *server_p = server->next;
             server = server->next;
-            destroy_yp_server (to_go);
+            destroy_yp_server(to_go);
             continue;
         }
         server_p = &server->next;
@@ -651,10 +688,9 @@ static void check_servers (void)
             ypdata_t *yp;
 
             source_t *source = node->key;
-            if (source->yp_public && (yp = create_yp_entry (source->mount)) != NULL)
+            if (source->yp_public && (yp = create_yp_entry (server, source->mount)) != NULL)
             {
                 ICECAST_LOG_DEBUG("Adding existing mount %s", source->mount);
-                yp->server = server;
                 yp->touch_interval = server->touch_interval;
                 yp->next = server->mounts;
                 server->mounts = yp;
@@ -689,7 +725,7 @@ static void add_pending_yp (struct yp_server *server)
 }
 
 
-static void delete_marked_yp (struct yp_server *server)
+static void delete_marked_yp(struct yp_server *server)
 {
     ypdata_t *yp = server->mounts, **prev = &server->mounts;
 
@@ -701,7 +737,7 @@ static void delete_marked_yp (struct yp_server *server)
             ICECAST_LOG_DEBUG("removed %s from YP server %s", yp->mount, server->url);
             *prev = yp->next;
             yp = yp->next;
-            yp_destroy_ypdata (to_go);
+            yp_destroy_ypdata(to_go);
             continue;
         }
         prev = &yp->next;
@@ -713,10 +749,12 @@ static void delete_marked_yp (struct yp_server *server)
 static void *yp_update_thread(void *arg)
 {
     ICECAST_LOG_INFO("YP update thread started");
+    int running;
 
     yp_running = 1;
-    while (yp_running)
-    {
+    running = 1;
+
+    while (running) {
         struct yp_server *server;
 
         thread_sleep (200000);
@@ -730,11 +768,10 @@ static void *yp_update_thread(void *arg)
             yp_process_server (server);
             server = server->next;
         }
-        thread_rwlock_unlock (&yp_lock);
-
         /* update the local YP structure */
         if (yp_update)
         {
+            thread_rwlock_unlock(&yp_lock);
             thread_rwlock_wlock (&yp_lock);
             check_servers ();
             server = (struct yp_server *)active_yps;
@@ -746,8 +783,9 @@ static void *yp_update_thread(void *arg)
                 server = server->next;
             }
             yp_update = 0;
-            thread_rwlock_unlock (&yp_lock);
         }
+        running = yp_running;
+        thread_rwlock_unlock(&yp_lock);
     }
     thread_rwlock_destroy (&yp_lock);
     thread_mutex_destroy (&yp_pending_lock);
@@ -768,10 +806,10 @@ static void yp_destroy_ypdata(ypdata_t *ypdata)
 {
     if (ypdata) {
         if (ypdata->mount) {
-            free (ypdata->mount);
+            free(ypdata->mount);
         }
         if (ypdata->url) {
-            free (ypdata->url);
+            free(ypdata->url);
         }
         if (ypdata->sid) {
             free(ypdata->sid);
@@ -803,13 +841,13 @@ static void yp_destroy_ypdata(ypdata_t *ypdata)
         if (ypdata->audio_info) {
             free(ypdata->audio_info);
         }
-        free (ypdata->subtype);
-        free (ypdata->error_msg);
-        free (ypdata);
+        free(ypdata->subtype);
+        free(ypdata->error_msg);
+        free(ypdata);
     }
 }
 
-static void add_yp_info (ypdata_t *yp, void *info, int type)
+static void add_yp_info(ypdata_t *yp, void *info, int type)
 {
     char *escaped;
 
@@ -888,7 +926,7 @@ void yp_add (const char *mount)
         if (yp == NULL)
         {
             /* add new ypdata to each servers pending yp */
-            yp = create_yp_entry (mount);
+            yp = create_yp_entry (server, mount);
             if (yp)
             {
                 ICECAST_LOG_DEBUG("Adding %s to %s", mount, server->url);
@@ -976,11 +1014,13 @@ void yp_touch (const char *mount)
 
 void yp_shutdown (void)
 {
+    thread_rwlock_wlock(&yp_lock);
     yp_running = 0;
     yp_update = 1;
+    thread_rwlock_unlock(&yp_lock);
+
     if (yp_thread)
         thread_join (yp_thread);
-    curl_global_cleanup();
     free ((char*)server_version);
     server_version = NULL;
     ICECAST_LOG_INFO("YP thread down");
